/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

The PCX file format is the copyright of ZSoft.  Some areas of this module are
based on original ZSoft file specifications.

*****************************************************************************/

#include <pandora/main.h>
#include <pandora/system/all.h>
#include <pandora/graphics/all.h>
#include <pandora/files/file.h>

#define VER_PCXMod 1.0

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

MODULE_HEADER = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_PCXMod, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 1998-2002.  All rights reserved.",
   "November 2002",
   "PCX"
};

static struct KernelBase *KernelBase = 0;
static OBJECTPTR PCXClass = 0;

static ERROR PCX_Activate(struct Picture *, APTR);
static ERROR PCX_Init(struct Picture *, APTR);
static ERROR PCX_Query(struct Picture *, APTR);
static ERROR PCX_SaveToObject(struct Picture *, struct acSaveToObject *);

static struct ActionArray PCXActions[] = {
   { AC_Activate,     PCX_Activate },
   { AC_Init,         PCX_Init },
   { AC_Query,        PCX_Query },
   { AC_SaveToObject, PCX_SaveToObject },
   { NULL, NULL }
};

struct PCXHeader {
   BYTE  Signature;      /* [00] Always set to 10 */
   BYTE  Version;        /* [01] Version 5 is the latest */
   BYTE  Encoding;       /* [02] Set to 1 if encoding is on */
   BYTE  BitsPerPixel;   /* [03] 1, 2, 4, 8 */
   WORD  XMin, YMin;     /* [04] Window dimensions */
   WORD  XMax, YMax;     /* [08] */
   WORD  XDPI, YDPI;     /* [12] Horizontal Dots Per Inch */
   UBYTE Palette[48];    /* [16] Palette information */
   BYTE  Reserved;       /* [64] NULL */
   BYTE  NPlanes;        /* [65] Number of Planes */
   WORD  BytesPerLine;   /* [66] Bytes per line */
   WORD  PaletteInfo;    /* [68] */
   WORD  HRes, VRes;     /* [70] */
   UBYTE Dummy[54];      /* [74] Filler to reach 128 bytes */
};

/*****************************************************************************
** Command: Init()
*/

static ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&PCXClass, NULL,
      FID_BaseClassID|TLONG,       ID_PICTURE,
      FID_SubClassID|TLONG,        ID_PCX,
      FID_Name|TSTRING,            "PCX",
      FID_Category|TLONG,          CCF_GRAPHICS,
      FID_FileExtension|TSTRING,   "pcx",
      FID_FileDescription|TSTRING, "PCX Picture",
      FID_Actions|TPTR,            PCXActions,
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

ERROR CMDExpunge(void)
{
   if (PCXClass) { Action(AC_Free, PCXClass, NULL); PCXClass = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************
** Function: UnpackRow()
** Synopsis: UBYTE * UnpackRow(struct PCXHeader *, UBYTE *Input, UBYTE *Output)
** Short:    Unpacks a single line of PCX data.
*/

static UBYTE * UnpackRow(struct PCXHeader *PCX, UBYTE *Input, UBYTE *Output)
{
   WORD x, i, cnt, TotalBytes;

   TotalBytes = PCX->BytesPerLine * PCX->NPlanes;
   if (PCX->Encoding IS 1) {
      x = 0;
      while (x < TotalBytes) {
         if ((0xc0 & *Input) IS 0xc0) { /* If the top two bits are set, the remaining */
            cnt = 0x3f & *Input++;      /* bits specify how many bytes to unpack */
            for (i = 0; i < cnt; i++) {
               *Output++ = *Input;
               if (x >= TotalBytes) {
                  DPrintF("!PCX:","Buffer overrun.");
                  return(Input);
               }
            }
            Input++;
            x += cnt;
         }
         else {                         /* Otherwise just output 1 byte */
            *Output++ = *Input++;
            x++;
         }
      }
      return(Input);
   }
   else {
      for (x = 0; x < TotalBytes; x++) {
         Output[x] = Input[x];
      }
      return(Input + x);
   }
}

/*****************************************************************************
** Picture: Activate
*/

static ERROR PCX_Activate(struct Picture *Self, APTR Void)
{
   struct PCXHeader PCX;
   struct Bitmap *Bitmap;
   struct acRead read;
   struct mtDrawPixel pixel;
   struct mtDrawRGBPixel drawrgb;
   struct RGB rgb;
   UBYTE *PCXData, *Buffer;
   WORD  y, MaxWidth, ydest, XRemainder, YRemainder, j, i, Domain;
   WORD SrcHeight, SrcWidth, error, Height;
   LONG DataSize, TotalBytes;
   STRING Location;

   DPrintF("Activate()","[PCX:%d]", Self->Head.UniqueID);

   PCXData = NULL;
   Buffer  = NULL;

   /*** Query the picture file first to retrieve bitmap and palette info ***/

   if ((error = Action(AC_Query, Self, NULL)) != ERR_Okay) {
      return(error);
   }

   /*** Open the file ***/

   if (!Self->prvFile) {
      if (GetField(Self, FID_Location, FT_STRING, &Location) IS ERR_Okay) {
         if (CreateObject(ID_FILE, NULL, &Self->prvFile, NULL,
            FID_Flags|TLONG,      FL_READ|FL_APPROXIMATE,
            FID_Location|TSTRING, Location,
            TAGEND) != ERR_Okay) return(ObjectError(Self, ERH_Activate, ERR_CreateObject));
      }
      else return(ObjectError(Self, ERH_Activate, ERR_GetField));
   }

   GetField(Self->prvFile, FID_Size, FT_LONG, &DataSize);
   DataSize -= sizeof(struct PCXHeader);
   if (DataSize < 1) goto exit;

   ActionTags(AC_Seek, Self->prvFile, 0.0, SEEK_START);
   read.Buffer = (APTR)&PCX;
   read.Length = sizeof(struct PCXHeader);
   if ((Action(AC_Read, Self->prvFile, &read) IS ERR_Okay) AND (read.Result != sizeof(struct PCXHeader))) {
      return(ObjectError(Self, ERH_Activate, ERR_Read));
   }

   DPrintF("3PCX:","File details: %dx%dx%d", PCX.XMax+1, PCX.YMax+1, PCX.NPlanes);

   /*** Load the PCX data ***/

   if (AllocMemory(DataSize, MEM_DATA|MEM_NOCLEAR, (APTR *)&PCXData, NULL) IS ERR_Okay) {
      SetRead(read, PCXData, DataSize);
      Action(AC_Read, Self->prvFile, &read);
      if (read.Result != DataSize) {
         FreeMemory(PCXData);
         return(ObjectError(Self, ERH_Activate, ERR_Read));
      }
   }
   else return(ObjectError(Self, ERH_Activate, ERR_Memory));

   /*** Initialise the Bitmap ***/

   if (Action(AC_Init, Self->Bitmap, NULL) != ERR_Okay) return(ERR_Init);

   Bitmap = Self->Bitmap;

   /*** Standard PCX calculations ***/

   SrcWidth   = PCX.XMax - PCX.XMin + 1;
   SrcHeight  = PCX.YMax - PCX.YMin + 1;
   TotalBytes = PCX.BytesPerLine * PCX.NPlanes;

   if (AllocMemory(TotalBytes, MEM_DATA|MEM_NOCLEAR, (APTR *)&Buffer, NULL) != ERR_Okay) goto exit;

   /*** Calculate dimensions for unpacking ***/

   if (Bitmap->Width < SrcWidth) MaxWidth = Bitmap->Width;
   else MaxWidth = SrcWidth;

   if (Bitmap->Height < SrcHeight) Height = Bitmap->Height;
   else Height = SrcHeight;

   Location   = PCXData;
   ydest      = 0;
   YRemainder = 0;

   if (Self->Flags & PCF_RESIZE) {
      DPrintF("Activate:","[PCX] Image will be resized to dimensions %dx%d.", Self->Bitmap->Width, Self->Bitmap->Height);
   }

   for (y=0; y < Height; y++) {
      Location = UnpackRow(&PCX, Location, Buffer);

      if (Location > (PCXData + DataSize)) {
         DPrintF("@Activate:","[PCX] Input overrun at line %d of %d!", y, Height);
         break;
      }

      if ((Self->Flags & PCF_RESIZEX) AND (Bitmap->Width != SrcWidth)) {
         if (SrcWidth > Bitmap->Width) { /*** Shrink Row ***/
            XRemainder = 0;
            Domain = SrcWidth - Bitmap->Width;
            j = 0;

            for (i=0; i < Bitmap->Width; i++) {
               XRemainder += Domain;
               while (XRemainder >= Bitmap->Width) {
                  XRemainder -= Bitmap->Width;
                  j++;
               }
               if (PCX.NPlanes IS 1) Bitmap->DrawUCPixel(Bitmap, i, ydest, Buffer[i]);
               else Bitmap->DrawUCPixel(Bitmap, i, ydest, PackPixel(Bitmap, Buffer[i], Buffer[PCX.BytesPerLine + i], Buffer[PCX.BytesPerLine + PCX.BytesPerLine + i]));
            }
         }
         else { /*** Expand Row ***/
            XRemainder = 0;
            Domain = Bitmap->Width - SrcWidth;
            j = 0;
            drawrgb.YCoord = ydest;
            pixel.YCoord = ydest;

            for (i=0; i < SrcWidth; i++) {
               if (PCX.NPlanes IS 1) Bitmap->DrawUCPixel(Bitmap, j++, ydest, Buffer[i]);
               else Bitmap->DrawUCPixel(Bitmap, j++, ydest, PackPixel(Bitmap, Buffer[i], Buffer[PCX.BytesPerLine+i], Buffer[(PCX.BytesPerLine<<1)+i]));

               XRemainder += Domain;
               while (XRemainder >= SrcWidth) {
                  if (PCX.NPlanes IS 1) Bitmap->DrawUCPixel(Bitmap, j++, ydest, Buffer[i]);
                  else Bitmap->DrawUCPixel(Bitmap, j++, ydest, PackPixel(Bitmap, Buffer[i], Buffer[PCX.BytesPerLine+i], Buffer[(PCX.BytesPerLine<<1)+i]));
                  XRemainder -= SrcWidth;
               }
            }
         }
      }
      else if (PCX.NPlanes IS 1) {
         if (Bitmap->BitsPerPixel IS 8) {
            for (i=0; i < MaxWidth; i++) Bitmap->DrawUCPixel(Bitmap, i, ydest, Buffer[i]);
         }
         else for (i=0; i < MaxWidth; i++) {
            rgb.Red   = Bitmap->Palette->Col[Buffer[i]].Red;
            rgb.Green = Bitmap->Palette->Col[Buffer[i]].Green;
            rgb.Blue  = Bitmap->Palette->Col[Buffer[i]].Blue;
            rgb.Alpha = 0;
            Bitmap->DrawUCRPixel(Bitmap, i, ydest, &rgb);
         }
      }
      else if (Bitmap->BitsPerPixel >= 15) {
         for (i=0; i < MaxWidth; i++) {
            Bitmap->DrawUCPixel(Bitmap, i, ydest, PackPixel(Bitmap, Buffer[i], Buffer[PCX.BytesPerLine + i], Buffer[(PCX.BytesPerLine<<1) + i]));
         }
      }
      else for (i=0; i < MaxWidth; i++) {
         rgb.Red   = Buffer[i];
         rgb.Green = Buffer[PCX.BytesPerLine + i];
         rgb.Blue  = Buffer[(PCX.BytesPerLine<<1) + i];
         rgb.Alpha = 0;
         Bitmap->DrawUCRPixel(Bitmap, i, ydest, &rgb);
      }

      /*** Image Resizing Y Axis ***/

      if (Self->Flags & PCF_RESIZEY) {
         if (Bitmap->Height < SrcHeight) { /*** Shrink Down ***/
            Domain      = SrcHeight - Bitmap->Height;
            YRemainder += Domain;
            while (YRemainder >= Bitmap->Height) {
               YRemainder -= Bitmap->Height;
               Location = UnpackRow(&PCX, Location, Buffer);
            }
         }
         else if (Bitmap->Height > SrcHeight) { /*** Expand ***/
            Domain      = Bitmap->Height - SrcHeight;
            YRemainder += Domain;
            while (YRemainder >= SrcHeight) {
               YRemainder -= SrcHeight;
               ydest++;
               ActionTags(MT_BlitArea, Bitmap, Bitmap, NULL, 0, ydest-1,
                 Bitmap->Width, 1, 0, ydest);
            }
         }
      }
      ydest++;
   }

   /*** Set mask attributes ***/

   if (PCX.NPlanes IS 1) {
      if ((!Bitmap->ColourRGB->Red) AND (!Bitmap->ColourRGB->Green) AND (!Bitmap->ColourRGB->Blue)) {
         if (!Bitmap->ColourIndex) Bitmap->ColourIndex = 0;
         Bitmap->ColourRGB->Red   = Bitmap->Palette->Col[Bitmap->ColourIndex].Red;
         Bitmap->ColourRGB->Green = Bitmap->Palette->Col[Bitmap->ColourIndex].Green;
         Bitmap->ColourRGB->Blue  = Bitmap->Palette->Col[Bitmap->ColourIndex].Blue;
         Bitmap->ColourRGB->Alpha = 0;
      }
   }

   if (Self->Flags & PCF_CACHE) Action(MT_CacheBitmap, Self, NULL);

   error = ERR_Okay;

exit:
   if (PCXData) FreeMemory(PCXData);
   if (Buffer)  FreeMemory(Buffer);
   Action(AC_Free, Self->prvFile, NULL);
   Self->prvFile = NULL;
   return(error);
}

/*****************************************************************************
** Picture: Init
*/

static ERROR PCX_Init(struct Picture *Self, APTR Void)
{
   UBYTE *buffer;
   STRING location;

   GetField(Self, FID_Location, FT_STRING, &location);

   if ((!location) OR (Self->Flags & PCF_NEW)) {
      /* If no location has been specified, assume that the picture is being
      ** created from scratch (e.g. to save an image to disk).  The
      ** programmer is required to specify the dimensions and colours of the
      ** Bitmap so that we can initialise it.
      */

      if (!Self->Bitmap->Width) Self->Bitmap->Width = Self->DisplayWidth;
      if (!Self->Bitmap->Height) Self->Bitmap->Height = Self->DisplayHeight;

      if ((Self->Bitmap->Width) AND (Self->Bitmap->Height)) {
         if (Action(AC_Init, Self->Bitmap, NULL) IS ERR_Okay) {
            return(ERR_Okay);
         }
         else return(ObjectError(Self, ERH_Init, ERR_Init));
      }
      else return(ObjectError(Self, ERH_Init, ERR_FieldNotSet));
   }
   else if (GetField(Self, FID_Header, FT_POINTER, &buffer) IS ERR_Okay) {
      if ((buffer[0] IS 0x0a) AND (buffer[1] IS 0x05)) {
         DPrintF("2PCX:","File \"%s\" is in PCX format.", location);
         if (!(Self->Flags & PCF_QUERY)) Action(AC_Activate, Self, NULL);
         return(ERR_Okay);
      }
   }

   return(ERR_NoSupport);
}

/*****************************************************************************
** Picture: Query
*/

static ERROR PCX_Query(struct Picture *Self, APTR Void)
{
   struct Bitmap    *Bitmap;
   struct PCXHeader *PCXHeader;
   struct acSeek    SeekArgs;
   struct acRead    read;
   STRING Location;
   LONG error;
   WORD i;

   error     = ERR_Okay;
   PCXHeader = NULL;

   if (!(Bitmap = Self->Bitmap)) return(ERR_ObjectCorrupt);

   if (!Self->prvFile) {
      GetField(Self, FID_Location, FT_STRING, &Location);
      if (GetField(Self, FID_Location, FT_STRING, &Location) IS ERR_Okay) {
         if (CreateObject(ID_FILE, NULL, &Self->prvFile, NULL,
            FID_Flags|TLONG,      FL_READ|FL_APPROXIMATE,
            FID_Location|TSTRING, Location,
            TAGEND) IS ERR_Okay) {
         }
         else return(ObjectError(Self, ERH_Activate, ERR_CreateObject));
      }
      else return(ObjectError(Self, ERH_Activate, ERR_GetField));
   }

   ActionTags(AC_Seek, Self->prvFile, 0.0, SEEK_START);
   if (AllocMemory(sizeof(struct PCXHeader), MEM_DATA|MEM_NOCLEAR, (APTR *)&PCXHeader, NULL) IS ERR_Okay) {
      SetRead(read, (APTR)PCXHeader, sizeof(struct PCXHeader));
      Action(AC_Read, Self->prvFile, &read);

      if (read.Result != sizeof(struct PCXHeader)) error = ObjectError(Self, ERH_Query, ERR_Read);
   }
   else error = ObjectError(Self, ERH_Query, ERR_Memory);

   if (error IS ERR_Okay) {
      if (!Bitmap->Width)       Bitmap->Width       = PCXHeader->XMax + 1;
      if (!Bitmap->Height)      Bitmap->Height      = PCXHeader->YMax + 1;
      if (!Self->DisplayWidth)  Self->DisplayWidth  = Bitmap->Width;
      if (!Self->DisplayHeight) Self->DisplayHeight = Bitmap->Height;
      if (!Bitmap->Type)        Bitmap->Type        = BMP_CHUNKY;
      if (!Bitmap->BitsPerPixel) {
         if (PCXHeader->NPlanes >= 3) Bitmap->BitsPerPixel = 24;
         else Bitmap->BitsPerPixel = 8;
      }

      if ((PCXHeader->NPlanes IS 1) AND (PCXHeader->Version IS 5)) {
         UBYTE Palette[(256*3)+1];

         Bitmap->Palette->AmtColours  = 256;

         SetSeek(SeekArgs, POS_END, 256 * 3);
         Action(AC_Seek, Self->prvFile, &SeekArgs);

         SetRead(read, Palette, 256 * 3);
         Action(AC_Read, Self->prvFile, &read);

         for (i=0; i < 256; i++) {
            Bitmap->Palette->Col[i].Red   = Palette[i*3];
            Bitmap->Palette->Col[i].Green = Palette[(i*3)+1];
            Bitmap->Palette->Col[i].Blue  = Palette[(i*3)+2];
            Bitmap->Palette->Col[i].Alpha = 0;
         }
      }

      /*** Re-query the Self->Bitmap ***/

      error = Action(AC_Query, Bitmap, NULL);
   }

   if (PCXHeader) FreeMemory(PCXHeader);
   return(error);
}

/*****************************************************************************
** PCX: SaveToObject()
**
** Please note that this routine only supports saving of 256 colour files
** at the moment.
*/

static ERROR PCX_SaveToObject(struct Picture *Self, struct acSaveToObject *Args)
{
   struct acWrite WriteArgs;
   struct mtReadPixel RPArgs;
   struct PCXHeader *pcx;
   struct Bitmap    *Bitmap;
   OBJECTPTR DestObject;
   UBYTE *buffer, *abuffer;
   UBYTE lastpixel, newpixel, counter;
   LONG  dp, i, j, size;

   if (AccessObject(Args->DestinationID, 1000, &DestObject) != ERR_Okay) {
      return(ObjectError(Self, ERH_SaveToObject, ERR_ExclusiveDenied));
   }

   Bitmap  = Self->Bitmap;
   dp      = 0;
   counter = 1;

   if (AllocMemory(sizeof(struct PCXHeader), MEM_DATA, (APTR *)&pcx, NULL) IS ERR_Okay) {
      pcx->Signature    = 10;      /* ZSoft PCX-files */
      pcx->Version      = 5;       /* Version */
      pcx->Encoding     = 1;       /* Run Length Encoding=ON */
      pcx->BitsPerPixel = 8;
      pcx->XMin = 0;
      pcx->YMin = 0;

      i = Bitmap->Width - 1;
      pcx->XMax = (i-((i>>8) * 256)) * 256 + (i>>8); /* Roll word */

      i = Bitmap->Height - 1;
      pcx->YMax = (i-((i>>8) * 256)) * 256 + (i>>8); /* Roll word */

      i = Bitmap->Width;
      pcx->HRes = (i-((i>>8) * 256)) * 256 + (i>>8); /* Roll word */

      i = Bitmap->Height;
      pcx->VRes = (i-((i>>8) * 256)) * 256 + (i>>8); /* Roll word */

      pcx->Reserved  = 0;
      pcx->NPlanes   = 1;   /* Bitmap->Planes */
      i = Bitmap->Width;
      pcx->BytesPerLine = (i - ((i>>8) * 256)) * 256 + (i>>8); /* Roll word */
      i = 1;
      pcx->PaletteInfo = (i - ((i>>8) * 256)) * 256 + (i>>8); /* Roll word */

      /*** Write header ***/

      SetWrite(WriteArgs, (APTR)pcx, sizeof(struct PCXHeader));
      Action(AC_Write, DestObject, &WriteArgs);
      FreeMemory(pcx);
   }
   else {
      ReleaseObject(DestObject);
      return(ObjectError(Self, ERH_SaveToObject, ERR_Memory));
   }

   /*** Setup body ***/

   size = Bitmap->Width * Bitmap->Height;    /* CHUNKY Size */

   if (AllocMemory(size, MEM_DATA|MEM_NOCLEAR, (APTR *)&buffer, NULL) IS ERR_Okay) {
      for (i=0; i <= Bitmap->Height-1; i++) {
         RPArgs.XCoord  = 0;
         RPArgs.YCoord  = i;
         Action(MT_ReadPixel, Bitmap, &RPArgs);
         lastpixel = RPArgs.Colour;

         for (j=1; j <= Bitmap->Width;j++) {
            RPArgs.XCoord  = j;
            RPArgs.YCoord  = i;
            Action(MT_ReadPixel, Bitmap, &RPArgs);
            newpixel = RPArgs.Colour;

            if ((newpixel IS lastpixel) AND (j != Bitmap->Width-1) AND (counter <= 62)) {
               counter++;
            }
            else {
               if (!(counter IS 1 AND lastpixel<192)) {
                  buffer[dp++] = 192 + counter;
               }
               buffer[dp++] = lastpixel;
               lastpixel = newpixel;
               counter = 1;
            }

            if (dp >= size-10) {
               FreeMemory(buffer);
               ReleaseObject(DestObject);
               return(ObjectError(Self, ERH_SaveToObject, ERR_Failed)); /* Buffer overflow */
            }
         }
      }
      SetWrite(WriteArgs, buffer, dp);
      Action(AC_Write, DestObject, &WriteArgs);
      FreeMemory(buffer);
   }
   else {
      ReleaseObject(DestObject);
      return(ObjectError(Self, ERH_SaveToObject, ERR_Failed));
   }

   /*** Setup palette ***/

   if (AllocMemory((256 * 3) + 1, MEM_DATA|MEM_NOCLEAR, (APTR *)&buffer, NULL) IS ERR_Okay) {
      abuffer = buffer;
      *buffer++ = 12;

      for (i=0; i < 256; i++) {
         *buffer++ = Bitmap->Palette->Col[i].Red;
         *buffer++ = Bitmap->Palette->Col[i].Green;
         *buffer++ = Bitmap->Palette->Col[i].Blue;
         j++;
      }

      SetWrite(WriteArgs, abuffer, (256 * 3) + 1);
      Action(AC_Write, DestObject, &WriteArgs);
      FreeMemory(abuffer);
   }
   else {
      ReleaseObject(DestObject);
      return(ObjectError(Self, ERH_SaveToObject, ERR_Memory));
   }

   ReleaseObject(DestObject);
   return(ERR_Okay);
}

