/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      Text
Version:   2.0
ID:        ID_TEXT
Status:    Stable
Include:   pandora/graphics/text.h
Category:  GUI
Date:      January 2004
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 1999-2004.  All rights reserved.
Keywords:  text, list, editor, string, write, document, edit, font
Short:     Provides text display and editing facilities.

###DESCRIPTION###
<p>The Text class provides a complete text display and editing service that
is suitable for almost any situation that requires effective text
management.  The class is most effective when applied to general text
display, text editing services, command-lines and input boxes.</p>

<p>The Text class is closely linked to the Font class, which provides all of
the code necessary for font management.  If you require information on how
to set font definitions such as the face and colour of the font, please
refer to the documentation for the Font class.  All fields in the Font class
are inherited, and you will find that the Face, Colour, Bold, Point and
Align fields are particularly helpful for setting font attributes.</p>

<p>The graphical area of a text object is defined using the standard
dimension conventions (x, y, width and height).  Margins (left, right, top
and bottom) also allow you to offset the text from the edges of the drawable
area.  The Text class draws its graphics to the foreground only, so you have
the choice of defining your own background to be placed behind the text.  If
you wish to forgo that in favour of a clear background, set the Background
field to your preferred colour.</p>

<p>The following example shows how to create a simple string display:</p>

<pre>
   &lt;text string=&quot;Hello World&quot; x=&quot;5&quot; y=&quot;10&quot; colour=&quot;#303030&quot;/&gt;
</pre>

<p>If you intend to create a text object that accepts user input, there are
a number of flags available to you that decide how the object will behave
as the user enters text information.  You may also attach child objects
that can be activated when the user presses the enter key after typing in
some information.  This can be useful for creating a custom-built reaction
to user input.  Here is an example of a text object that runs a script when
the enter key is pressed:</p>

<pre>
   &lt;text flags=&quot;commandline&quot; x=&quot;3&quot; y=&quot;4&quot;
     face=&quot;[gl_fonts.face]&quot; colour=&quot;#000000&quot;&gt;
      &lt;script static src=&quot;myscript.dml&quot; &amp;string=&quot;$[owner.string]$&quot;/&gt;
   &lt;/text&gt;
</pre>

<p>If the text object needs to be linked to another object, you may prefer
to set the SelectionObject and SelectionField values so that they point to an
external object field.  This ensures that any information typed into the
text object is always reflected in the field of the linked object.</p>

<p>For long text lists, scrollbars can be attached via the HScroll and VScroll
fields.  For hints on how to use a text object to build a full featured text
editing application, refer to the script file located at
&quot;scripts:applications/text/editor/main.dml&quot;.</p>

<p>By default the Text class supports text highlighting for cut, copy and
paste operations.  This support is backed by system keypresses such as
CTRL-C, CTRL-V and CTRL-X.  If you need to send custom commands to the Text
class that aren't covered by the control keys, use the <a
href="keypress.html">Keypress</a> class to create response mechanisms for
the keyboard.</p>

###END###

*****************************************************************************/

#include <pandora/main.h>
#include <pandora/data/xml.h>
#include <pandora/files/file.h>
#include <pandora/graphics/bitmap.h>
#include <pandora/graphics/font.h>
#include <pandora/graphics/render.h>
#include <pandora/graphics/text.h>
#include <pandora/input/clipboard.h>
#include <pandora/input/keyboard.h>
#include <pandora/input/joyport.h>
#include <pandora/graphics/scroll.h>
#include <pandora/modules/pointer.h>
#include <pandora/modules/strings.h>
#include <pandora/modules/render.h>
#include <pandora/system/all.h>
#include <pandora/functions/set.c>

//#define DBG_TEXT TRUE

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_TEXT 1.0

MODULE_HEADER = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_TEXT, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2001-2004.  All rights reserved.",
   "January 2004",
   "Text"
};

static struct KernelBase  *KernelBase;
static struct StringsBase *StringsBase;
static struct RenderBase  *RenderBase;
static struct PointerBase *PointerBase;
static OBJECTPTR TextClass = 0;
static OBJECTPTR modStrings = NULL, modRender = NULL, modPointer = NULL;
static OBJECTID glClipboardID;
static struct RGB glHighlight = { 220, 220, 255 };

/*****************************************************************************
** Text class definition.
*/

#define COLOUR_LENGTH 16
#define CURSOR_RATE 1400

struct TextHistory {
   LONG  Number;
   UBYTE Buffer[120];
};

struct Text {
   OBJECT_HEADER            /* Class header */
   OBJECTID VScroll;        /* Reference to a vertical scroll bar */
   OBJECTID HScroll;        /* References to a horizontal scroll bar */
   OBJECTID TabFocusID;     /* If the tab key is pressed, a focus can be sent to this object */
   OBJECTID DrawableID;     /* The object that should be rendered to */
   OBJECTID InputID;        /* The keyboard to read information from */
   OBJECTID FocusID;        /* */
   OBJECTID SelectionObjectID; /* */
   LONG     CursorColumn;   /* Cursor column */
   LONG     CursorRow;      /* Cursor row / Line selection */
   LONG     Flags;          /* Special flags */
   LONG     AmtLines;       /* Amount of strings currently in the array */
   LONG     SelectRow;      /* Row that the current text selection starts from */
   LONG     SelectColumn;   /* Column that the current text selection starts from */
   struct Font *Font;       /* Font to use for the text */
   LONG     LeftMargin;     /* Size of the left margin */
   LONG     TopMargin;      /* Size of the top margin */
   LONG     RightMargin;    /* Size of the right margin */
   LONG     BottomMargin;   /* Size of the bottom margin */
   LONG     Effect;         /* */
   LONG     Frame;          /* The frame that the Text will appear on */
   LONG     HistorySize;    /* Size of the history buffer (measured in array elements) */
   LONG     LineLimit;      /* Limit the number of lines to the value in this field */
   LONG     CharLimit;      /* Limit the number of chars allowed in each line of the text object */
   LONG     Visible;        /* True/False */
   LONG     EffectDelay;    /* If the effect should be delayed, specify the time in microseconds */
   LONG     Duration;
   LONG     Activation;

   PRIVATE_FIELDS
   struct RGB Highlight;    /* Default colour for text highlighting */
   struct RGB Background;   /* Colour for text background */
   struct RGB CursorRGB;    /* The colour of the cursor */
   struct TextLine *Array;  /* Pointer to an array of lines */
   struct TextHistory *History; /* History of text entries */
   struct Font *BoldFont;   /* Bold version of the current font */
   struct Font *ItalicFont; /* Italic version of the current font */
   struct Font *BIFont;     /* Bold-Italic version of the current font */
   OBJECTPTR XML;
   BYTE   SelectionField[32]; /* When the focus is lost, the string is sent to this field */
   WORD   *Frames;
   FLOAT  XCoord;            /* Dimensions for the text drawable area */
   FLOAT  YCoord;
   FLOAT  XOffset;
   FLOAT  YOffset;
   FLOAT  Width;
   FLOAT  Height;
   LONG   DrawableWidth;
   LONG   DrawableHeight;
   LONG   CurrentFrame;
   LONG   MaxLines;          /* Total number of lines available in the array */
   LONG   Dimensions;        /* Dimension flags */
   LONG   XPosition;         /* Horizontal scrolling offset */
   LONG   YPosition;         /* Vertical scrolling offset */
   LONG   CursorChar;
   LONG   AreaX, AreaY, AreaWidth, AreaHeight;
   STRING Location;         /* Optional file to load on Init() */
   WORD   KeyboardChannel;  /* Set to true if keyboard channel is active */
   WORD   TotalFrames;
   WORD   ClickHeld;
   WORD   HistoryPos;
   WORD   CursorWidth;
   UBYTE  PointerLocked;
   UBYTE  State;
   BYTE   Tag;
   BYTE   NoUpdate;
   BYTE   NoCursor;
   UWORD  CursorFlash;     /* If set to 1, cursor is invisible.  This field is managed by the Timer action */
   WORD   CharCount;       /* Used by the Typist special effect */
   struct RGB prvFlash;    /* Current colour used by the flash effect */
   WORD   FlashCount;
   WORD   EffectActive;
   LONG   CursorSavePos;
   WORD   CursorEndColumn;
};

struct TextLine {
   STRING String;
   LONG   Length;
   LONG   PixelLength;
   UBYTE  *Tags;       /* Special tags affecting this line */
};

enum {
   STATE_ENTERED=1,
   STATE_INSIDE,
   STATE_EXITED
};

static ERROR GET_Height(struct Text *, struct Variable *);
static ERROR GET_Location(struct Text *, STRING *);
static ERROR GET_String(struct Text *, STRING *);
static ERROR GET_TextHeight(struct Text *, LONG *);
static ERROR GET_TextWidth(struct Text *, LONG *);
static ERROR GET_Width(struct Text *, struct Variable *);
static ERROR GET_X(struct Text *, struct Variable *);
static ERROR GET_XOffset(struct Text *, struct Variable *);
static ERROR GET_Y(struct Text *, struct Variable *);
static ERROR GET_YOffset(struct Text *, struct Variable *);

static ERROR SET_Background(struct Text *, STRING);
static ERROR SET_CharLimit(struct Text *, LONG *);
static ERROR SET_CursorColumn(struct Text *, LONG *);
static ERROR SET_CursorRow(struct Text *, LONG *);
static ERROR SET_Flags(struct Text *, LONG *);
static ERROR SET_Frames(struct Text *, STRING);
static ERROR SET_Highlight(struct Text *, STRING);
static ERROR SET_HScroll(struct Text *, OBJECTID *);
static ERROR SET_Location(struct Text *, STRING);
static ERROR SET_Origin(struct Text *, STRING);
static ERROR SET_SelectionField(struct Text *, STRING);
static ERROR SET_String(struct Text *, STRING);
static ERROR SET_VScroll(struct Text *, OBJECTID *);
static ERROR SET_Height(struct Text *, struct Variable *);
static ERROR SET_Width(struct Text *, struct Variable *);
static ERROR SET_X(struct Text *, struct Variable *);
static ERROR SET_XOffset(struct Text *, struct Variable *);
static ERROR SET_Y(struct Text *, struct Variable *);
static ERROR SET_YOffset(struct Text *, struct Variable *);

static struct FieldDef TextFlags[] = {
   { "EDIT",          TXF_EDIT          }, { "ENTERACTIVATE", TXF_ENTERACTIVATE },
   { "AUTOCLEAR",     TXF_AUTOCLEAR     }, { "COMMANDLINE",   TXF_COMMANDLINE   },
   { "OVERWRITE",     TXF_OVERWRITE     }, { "FOCUSACTIVATE", TXF_FOCUSACTIVATE },
   { "WORDWRAP",      TXF_WORDWRAP      }, { "HISTORY",       TXF_HISTORY       },
   { "VARIABLE",      TXF_VARIABLE      }, { "EFFECTLOOP",    TXF_EFFECTLOOP    },
   { "NOFXDRAW",      TXF_NOFXDRAW      }, { "FORCECAPS",     TXF_FORCECAPS     },
   { "GLOBALEDITING", TXF_GLOBALEDITING }, { "AREASELECTED",  TXF_AREASELECTED  },
   { "NOSYSKEYS",     TXF_NOSYSKEYS     }, { "MULTISELECT",   TXF_MULTISELECT   },
   { "SINGLESELECT",  TXF_SINGLESELECT  }, { "DISABLED",      TXF_DISABLED      },
   { "BACKGROUND",    TXF_BACKGROUND    }, { "TABKEY",        TXF_TABKEY        },
   { "REPEATEFFECT",  TXF_EFFECTLOOP    }, { "PRESERVEBKGD",  TXF_PRESERVEBKGD  },
   { "SECRET",        TXF_SECRET        }, { "ENTERTAB",      TXF_ENTERTAB },
   /*** Synonyms ***/
   { "PASSWORD",      TXF_PASSWORD      },
   { NULL, NULL }
};

static struct FieldDef ActivationFlags[] = {
   { "SINGLECLICK",  ACF_SINGLECLICK  }, { "DOUBLECLICK", ACF_DOUBLECLICK },
   { "RELEASE",      ACF_RELEASE      }, { "MOVEMENT",    ACF_MOVEMENT    },
   /*** Synonyms ***/
   { "CLICKRELEASE", ACF_RELEASE      }, { "SENSITIVE",   ACF_SENSITIVE   },
   { NULL, NULL }
};

static struct FieldDef TextEffects[] = {
   { "TYPIST", EFX_TYPIST },
   { "FLASH",  EFX_FLASH },
   { "FADE",   EFX_FADE },
   { NULL, NULL }
};

static struct FieldArray TextFields[] = {
   { "VScroll",        0, FDF_OBJECTID|FDF_RW,  ID_SCROLL, NULL, SET_VScroll },
   { "HScroll",        0, FDF_OBJECTID|FDF_RW,  ID_SCROLL, NULL, SET_HScroll },
   { "TabFocus",       0, FDF_OBJECTID|FDF_RW,  0,         NULL, NULL },
   { "Drawable",       0, FDF_OBJECTID|FDF_RW,  ID_RENDER, NULL, NULL },
   { "Input",          0, FDF_OBJECTID|FDF_RW,  0,         NULL, NULL },
   { "Focus",          0, FDF_OBJECTID|FDF_RI,  0,         NULL, NULL },
   { "SelectionObject",0, FDF_OBJECTID|FDF_RW,  0,         NULL, NULL },
   { "CursorColumn",   0, FDF_LONG|FDF_RW,      0,         NULL, SET_CursorColumn },
   { "CursorRow",      0, FDF_LONG|FDF_RW,      0,         NULL, SET_CursorRow },
   { "Flags",          0, FDF_LONGFLAGS|FDF_RI, (LONG)&TextFlags, NULL, SET_Flags },
   { "AmtLines",       0, FDF_LONG|FDF_R,       0,         NULL, NULL },
   { "SelectRow",      0, FDF_LONG|FDF_R,       0,         NULL, NULL },
   { "SelectColumn",   0, FDF_LONG|FDF_R,       0,         NULL, NULL },
   { "Font",           0, FDF_CHILD|FDF_R,      ID_FONT,   NULL, NULL },
   { "LeftMargin",     0, FDF_LONG|FDF_RI,      0,         NULL, NULL },
   { "TopMargin",      0, FDF_LONG|FDF_RI,      0,         NULL, NULL },
   { "RightMargin",    0, FDF_LONG|FDF_RI,      0,         NULL, NULL },
   { "BottomMargin",   0, FDF_LONG|FDF_RI,      0,         NULL, NULL },
   { "Effect",         0, FDF_LONG|FDF_LOOKUP|FDF_RI,      (LONG)&TextEffects, NULL, NULL },
   { "Frame",          0, FDF_LONG|FDF_RW,      0,         NULL, NULL },
   { "HistorySize",    0, FDF_LONG|FDF_RI,      0,         NULL, NULL },
   { "LineLimit",      0, FDF_LONG|FDF_RW,      0,         NULL, NULL },
   { "CharLimit",      0, FDF_LONG|FDF_RW,      0,         NULL, SET_CharLimit },
   { "Visible",        0, FDF_LONG|FDF_RW,      0,         NULL, NULL },
   { "EffectDelay",    0, FDF_LONG|FDF_RI,      0,         NULL, NULL },
   { "Duration",       0, FDF_LONG|FDF_RW,      0,         NULL, NULL },
   { "Activation",     0, FDF_LONGFLAGS|FDF_RW, (LONG)&ActivationFlags, NULL, NULL },
   /*** Virtual fields ***/
   { "Frames",         0, FDF_STRING|FDF_W,   0, NULL,               SET_Frames },
   { "Background",     0, FDF_STRING|FDF_W,   0, NULL,               SET_Background },
   { "Highlight",      0, FDF_STRING|FDF_W,   0, NULL,               SET_Highlight },
   { "Location",       0, FDF_STRING|FDF_RW,  0, GET_Location,       SET_Location },
   { "Origin",         0, FDF_STRING|FDF_RW,  0, GET_Location,       SET_Origin },
   { "SelectionField", 0, FDF_STRING|FDF_W,  0, NULL,               SET_SelectionField },
   { "Src",            0, FDF_STRING|FDF_RW,  0, GET_Location,       SET_Location },
   { "String",         0, FDF_STRING|FDF_RW,  0, GET_String,         SET_String },
   { "TextHeight",     0, FDF_LONG|FDF_R,     0, GET_TextHeight,     NULL },
   { "TextWidth",      0, FDF_LONG|FDF_R,     0, GET_TextWidth,      NULL },
   /*** Variable fields ***/
   { "Height",  0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Height,  SET_Height },
   { "Width",   0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Width,   SET_Width },
   { "X",       0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_X,       SET_X },
   { "XCoord",  0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_X,       SET_X },
   { "XOffset", 0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_XOffset, SET_XOffset },
   { "Y",       0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Y,       SET_Y },
   { "YCoord",  0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Y,       SET_Y },
   { "YOffset", 0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_YOffset, SET_YOffset },
   END_FIELD
};

/*****************************************************************************
** Class actions.
*/

static ERROR TEXT_ActionNotify(struct Text *, struct acActionNotify *);
static ERROR TEXT_Activate(struct Text *, APTR);
static ERROR TEXT_Clear(struct Text *, APTR);
static ERROR TEXT_Clipboard(struct Text *, struct acClipboard *);
static ERROR TEXT_ClosingTag(struct Text *, APTR);
static ERROR TEXT_DataChannel(struct Text *, struct acDataChannel *);
static ERROR TEXT_Disable(struct Text *, APTR);
static ERROR TEXT_Draw(struct Text *, APTR);
static ERROR TEXT_Enable(struct Text *, APTR);
static ERROR TEXT_Focus(struct Text *, APTR);
static ERROR TEXT_Free(struct Text *, APTR);
static ERROR TEXT_Hide(struct Text *, APTR);
static ERROR TEXT_Init(struct Text *, APTR);
static ERROR TEXT_LostFocus(struct Text *, APTR);
static ERROR TEXT_NewObject(struct Text *, APTR);
static ERROR TEXT_NewOwner(struct Text *, struct acNewOwner *);
static ERROR TEXT_OpeningTag(struct Text *, APTR);
static ERROR TEXT_Show(struct Text *, APTR);
static ERROR TEXT_SaveToObject(struct Text *, struct acSaveToObject *);
static ERROR TEXT_ScrollToPoint(struct Text *, struct acScrollToPoint *);
static ERROR TEXT_Timer(struct Text *, struct acTimer *);
static ERROR TEXT_UserClick(struct Text *, struct acUserClick *);
static ERROR TEXT_UserClickRelease(struct Text *, struct acUserClickRelease *);
static ERROR TEXT_UserMovement(struct Text *, struct acUserMovement *);

static struct ActionArray TextActions[] = {
   { AC_ActionNotify,     TEXT_ActionNotify },
   { AC_Activate,         TEXT_Activate },
   { AC_Clear,            TEXT_Clear },
   { AC_Clipboard,        TEXT_Clipboard },
   { AC_ClosingTag,       TEXT_ClosingTag },
   { AC_DataChannel,      TEXT_DataChannel },
   { AC_Disable,          TEXT_Disable },
   { AC_Draw,             TEXT_Draw },
   { AC_Enable,           TEXT_Enable },
   { AC_Focus,            TEXT_Focus },
   { AC_Free,             TEXT_Free },
   { AC_Hide,             TEXT_Hide },
   { AC_Init,             TEXT_Init },
   { AC_LostFocus,        TEXT_LostFocus },
   { AC_NewOwner,         TEXT_NewOwner },
   { AC_NewObject,        TEXT_NewObject },
   { AC_OpeningTag,       TEXT_OpeningTag },
   { AC_SaveToObject,     TEXT_SaveToObject },
   { AC_ScrollToPoint,    TEXT_ScrollToPoint },
   { AC_Show,             TEXT_Show },
   { AC_Timer,            TEXT_Timer },
   { AC_UserClick,        TEXT_UserClick },
   { AC_UserClickRelease, TEXT_UserClickRelease },
   { AC_UserMovement,     TEXT_UserMovement },
   { NULL, NULL }
};

/*****************************************************************************
** Class methods.
*/

#define MTX_AddLine MT_AddTextLine
#define MTX_DeleteLine MT_DeleteTextLine
#define MTX_FindText MT_FindText
#define MTX_GetLine MT_GetTextLine
#define MTX_ReplaceLine MT_ReplaceTextLine
#define MTX_ReplaceText MT_ReplaceText
#define MTX_SelectTextArea MT_SelectTextArea
#define MTX_SetFont MT_SetTextFont

#define mtxDeleteLine mtDeleteTextLine
#define mtxAddLine mtAddTextLine
#define mtxFindText mtFindText
#define mtxGetLine mtGetTextLine
#define mtxReplaceLine mtReplaceTextLine
#define mtxReplaceText mtReplaceText
#define mtxSelectTextArea mtSelectTextArea
#define mtxSetFont mtSetTextFont

static ERROR TEXT_DeleteLine(struct Text *, struct mtxDeleteLine *);
static ERROR TEXT_AddLine(struct Text *, struct mtxAddLine *);
static ERROR TEXT_FindText(struct Text *, struct mtxFindText *);
static ERROR TEXT_GetLine(struct Text *, struct mtxGetLine *);
static ERROR TEXT_ReplaceLine(struct Text *, struct mtxReplaceLine *);
static ERROR TEXT_ReplaceText(struct Text *, struct mtxReplaceText *);
static ERROR TEXT_SelectTextArea(struct Text *, struct mtxSelectTextArea *);
static ERROR TEXT_SetFont(struct Text *, struct mtxSetFont *);

static struct FunctionField argsAddLine[]        = { { "String", ARG_STR }, { "Line", ARG_LONG }, { "Length", ARG_LONG }, { NULL, NULL } };
static struct FunctionField argsDeleteLine[]     = { { "Line", ARG_LONG }, { NULL, NULL } };
static struct FunctionField argsFindText[]       = { { "String", ARG_STR }, { "Flags", ARG_LONG }, { "Row", ARG_LONG|ARG_RESULT }, { "Column", ARG_LONG|ARG_RESULT }, { NULL, NULL } };
static struct FunctionField argsGetLine[]        = { { "Line", ARG_LONG }, { "Buffer", ARG_PTR }, { "Length", ARG_PTRSIZE }, { NULL, NULL } };
static struct FunctionField argsReplaceLine[]    = { { "String", ARG_STR }, { "Line", ARG_LONG }, { "Length", ARG_LONG }, { NULL, NULL } };
static struct FunctionField argsReplaceText[]    = { { "Find", ARG_STR }, { "Replace", ARG_STR }, { "Flags", ARG_LONG }, { NULL, NULL } };
static struct FunctionField argsSelectTextArea[] = { { "Row", ARG_LONG }, { "Column", ARG_LONG }, { "EndRow", ARG_LONG }, { "EndColumn", ARG_LONG }, { NULL, NULL } };
static struct FunctionField argsSetFont[]        = { { "Face", ARG_STR  }, { NULL, NULL } };

static struct MethodArray TextMethods[] = {
   { MT_AddTextLine,     TEXT_AddLine,        "AddLine",     argsAddLine,        sizeof(struct mtxAddLine) },
   { MT_DeleteTextLine,  TEXT_DeleteLine,     "DeleteLine",  argsDeleteLine,     sizeof(struct mtxDeleteLine) },
   { MT_GetTextLine,     TEXT_GetLine,        "GetLine",     argsGetLine,        sizeof(struct mtxGetLine) },
   { MT_ReplaceTextLine, TEXT_ReplaceLine,    "ReplaceLine", argsReplaceLine,    sizeof(struct mtxReplaceLine) },
   { MT_SelectTextArea,  TEXT_SelectTextArea, "SelectArea",  argsSelectTextArea, sizeof(struct mtxSelectTextArea) },
   { MT_FindText,        TEXT_FindText,       "FindText",    argsFindText,       sizeof(struct mtxFindText) },
   { MT_ReplaceText,     TEXT_ReplaceText,    "ReplaceText", argsReplaceText,    sizeof(struct mtxReplaceText) },
   { MT_SetTextFont,     TEXT_SetFont,        "SetFont",     argsSetFont,        sizeof(struct mtxSetFont) },
   { NULL, NULL, NULL, NULL }
};

/*****************************************************************************
** Tags for the text lines.
*/

enum {
   TF_BOLD = 1,
   TF_ITALICS,
   TF_COLOUR,
   TF_UNDERLINE,
   TF_END
};

/*****************************************************************************
** Function prototypes.
*/

static void  AddHistory(struct Text *, STRING);
static ERROR AddLine(struct Text *, STRING, LONG, LONG, LONG);
static ERROR AddXML(struct Text *, struct XMLTag *, WORD, LONG);
static ERROR CalculateHScroll(struct Text *);
static LONG  CalculateWidth(struct Text *, STRING, LONG);
static ERROR CalculateVScroll(struct Text *);
static ERROR CheckTag(UBYTE *, LONG, STRING *);
static LONG  ColumnCoordinate(struct Text *, LONG, LONG);
static void  DeleteSelectedArea(struct Text *);
static void  DrawLines(struct Text *, LONG, LONG);
static ERROR EditKeypress(struct Text *, LONG, LONG);
static void  GetSelectedArea(struct Text *, LONG *, LONG *, LONG *, LONG *);
static ERROR InsertLineTag(struct Text *, LONG, LONG, STRING);
static void  MoveCursor(struct Text *, LONG, LONG);
static WORD  ReadByte(STRING);
static void  Redraw(struct Text *);
static void  RedrawCursor(struct Text *, LONG);
static void  RedrawLine(struct Text *, LONG);
static ERROR ReplaceLine(struct Text *, STRING, LONG, LONG);
static LONG  RowCoordinate(struct Text *, LONG);
static void  ValidateCursorPosition(struct Text *, LONG);
static LONG  ViewCursor(struct Text *);
static LONG  ViewSelection(struct Text *);
static LONG  XMLContentLength(struct XMLTag *);
static void  XMLExtractContent(struct XMLTag *, UBYTE *, LONG *, WORD);

#define RemoveCursor(a) RedrawCursor((a),FALSE)

/*****************************************************************************
** Command: Init()
*/

ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   OBJECTID objectid;
   OBJECTPTR object;
   UBYTE buffer[30], *str;

   KernelBase = argKernelBase;

   if (CreateObject(ID_MODULE, NULL, &modStrings, NULL,
      FID_Name|TSTRING,   "strings",
      FID_Version|TFLOAT, MODVERSION_STRINGS,
      TAGEND) IS ERR_Okay) {
      if (GetField(modStrings, FID_ModBase, FT_POINTER, &StringsBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   if (CreateObject(ID_MODULE, NULL, &modRender, NULL,
      FID_Name|TSTRING,   "render",
      FID_Version|TFLOAT, MODVERSION_RENDER,
      TAGEND) IS ERR_Okay) {
      if (GetField(modRender, FID_ModBase, FT_POINTER, &RenderBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   if (CreateObject(ID_MODULE, NULL, &modPointer, NULL,
      FID_Name|TSTRING,   "pointer",
      FID_Version|TFLOAT, MODVERSION_POINTER,
      TAGEND) IS ERR_Okay) {
      if (GetField(modPointer, FID_ModBase, FT_POINTER, &PointerBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   FastFindObject("SystemClipboard", NULL, &glClipboardID, 1, NULL);

   if (FastFindObject("gl_colours", NULL, &objectid, 1, NULL) IS ERR_Okay) {
      if (AccessObject(objectid, 4000, &object) IS ERR_Okay) {
         if (GetUnlistedField(object, "TextHighlight", buffer, sizeof(buffer)) IS ERR_Okay) {
            str = buffer;
            if (*str IS '#') {
               str++;
               glHighlight.Red   = ReadByte(str);
               if (*str) str++; if (*str) str++;
               glHighlight.Green = ReadByte(str);
               if (*str) str++; if (*str) str++;
               glHighlight.Blue  = ReadByte(str);
            }
            else {
               glHighlight.Red = (LONG)StrToInt(str);
               while ((*str) AND (*str != ',')) str++; if (*str) str++;
               glHighlight.Green = (LONG)StrToInt(str);
               while ((*str) AND (*str != ',')) str++; if (*str) str++;
               glHighlight.Blue = (LONG)StrToInt(str);
            }

         }
         ReleaseObject(object);
      }
   }

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&TextClass, NULL,
      FID_BaseClassID|TLONG, ID_TEXT,
      FID_Version|TFLOAT,    VER_TEXT,
      FID_Name|TSTRING,      "Text",
      FID_Category|TLONG,    CCF_GUI,
      FID_Flags|TLONG,       CLF_PROMOTECHILDREN|CLF_PRIVATEONLY,
      FID_Actions|TPTR,      TextActions,
      FID_Methods|TPTR,      TextMethods,
      FID_Fields|TPTR,       TextFields,
      FID_Size|TLONG,        sizeof(struct Text),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

ERROR CMDExpunge(void)
{
   if (modPointer) { acFree(modPointer); modPointer = NULL; }
   if (modStrings) { acFree(modStrings); modStrings = NULL; }
   if (modRender)  { acFree(modRender);  modRender  = NULL; }
   if (TextClass)  { acFree(TextClass);  TextClass = NULL; }
   return(ERR_Okay);
}

/****************************************************************************/

static UBYTE UCase(UBYTE Case)
{
  if ((Case >= 'a') AND (Case <= 'z')) Case -= 0x20;
  return(Case);
}

/*****************************************************************************
** Text: ActionNotify
*/

static ERROR TEXT_ActionNotify(struct Text *Self, struct acActionNotify *Args)
{
   if (!Args) return(ERR_Args);

   if (Args->Error != ERR_Okay) return(ERR_Okay);

   if (Args->ActionID IS AC_Disable) {
      Action(AC_Disable, Self, NULL);
   }
   else if (Args->ActionID IS AC_Enable) {
      Action(AC_Enable, Self, NULL);
   }
   else if (Args->ActionID IS AC_Focus) {
      OBJECTPTR input;

      Self->CursorFlash = 0;

      if (Self->KeyboardChannel IS FALSE) {
         if (AccessObject(Self->InputID, 5000, &input) IS ERR_Okay) {
            SubscribeChannel(input, Self);
            Self->KeyboardChannel = TRUE;
            RedrawCursor(Self, TRUE);

            ReleaseObject(input);
         }
         else ObjectError(Self, ERH_Focus, ERR_AccessObject);
      }
   }
   else if (Args->ActionID IS AC_LostFocus) {
      Action(AC_LostFocus, Self, NULL);
   }
   else if (Args->ActionID IS AC_Redimension) {
      struct Variable var;
      struct acRedimension *redimension;

      redimension = (struct acRedimension *)Args->Args;

      /*** Recompute our dimensions ***/

      var.Type = FD_LARGE;
      GET_X(Self, &var); Self->AreaX = var.Large;
      GET_Y(Self, &var); Self->AreaY = var.Large;
      GET_Width(Self, &var);  Self->AreaWidth  = var.Large;
      GET_Height(Self, &var); Self->AreaHeight = var.Large;

      if ((Self->Flags & TXF_WORDWRAP) AND (redimension->Width != Self->DrawableWidth)) {
         CalculateVScroll(Self);
      }
      Self->DrawableWidth = redimension->Width;
      Self->DrawableHeight = redimension->Height;
   }
   else if (Args->ActionID IS AC_UserClick) {
      Action(AC_UserClick, Self, (struct acUserClick *)Args->Args);
   }
   else if (Args->ActionID IS AC_UserClickRelease) {
      Action(AC_UserClickRelease, Self, (struct acUserClickRelease *)Args->Args);
   }
   else if (Args->ActionID IS AC_UserMovement) {
      struct acUserMovement *movement;
      LONG width, height, xcoord, ycoord;
      UBYTE inside;

      /*** Determine the current movement state (exit, enter, inside) ***/

      movement = (struct acUserMovement *)Args->Args;

      if (Self->Flags & TXF_EDIT) {
         inside = TRUE;
         if (movement->OverObject IS Self->DrawableID) {
            if (Self->Dimensions) {
               if (GetFields(Self, FID_Width|TLONG,  &width,
                                   FID_Height|TLONG, &height,
                                   FID_XCoord|TLONG, &xcoord,
                                   FID_YCoord|TLONG, &ycoord,
                                   TAGEND) IS ERR_Okay) {
                  if ((movement->XCoord < xcoord) OR (movement->YCoord < ycoord) OR
                      (movement->XCoord >= xcoord + width) OR (movement->YCoord >= ycoord + height)) {
                     inside = FALSE;
                  }
               }
            }
         }
         else inside = FALSE;

         if (inside) {
            if (Self->State IS STATE_ENTERED) Self->State = STATE_INSIDE;
            else if (Self->State != STATE_INSIDE) {
               Self->State = STATE_ENTERED;

               SetCursor(NULL, ACF_BUFFER, PTR_TEXT, NULL, Self->Head.UniqueID, NULL);
               Self->PointerLocked = TRUE;
            }
         }
         else {
            if (Self->State != STATE_EXITED) {
               Self->State = STATE_EXITED;
               RestoreCursor(PTR_DEFAULT, Self->Head.UniqueID);
               Self->PointerLocked = FALSE;
            }
         }
      }

      Action(AC_UserMovement, Self, movement);
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Internal: DrawText()
*/

static void DrawText(struct Text *Self, struct Render *Drawable, struct Bitmap *Bitmap)
{
   struct mtCalcStringWidth calc;
   struct Font *font, *currentfont;
   struct RGB basergb;
   STRING str;
   LONG i, x, sx, row, column, endrow, endcolumn, width;
   LONG style, selectrow, selectcolumn, alignoffset;
   LONG clipleft, clipright, cliptop, clipbottom, amtlines;
   LONG chars, len;
   WORD valign;

   if ((Self->Visible IS FALSE) OR (Self->Tag)) return;

   if (!(font = Self->Font)) return;

   /*** Frame testing ***/

   if ((Self->Frame) OR (Self->Frames)) {
      if (Self->Frame) {
         if (Drawable->Frame != Self->Frame) return;
      }
      else if (Self->Frames) {
         if ((Drawable->Frame > Self->TotalFrames) OR (Self->Frames[Drawable->Frame] < 1)) return;
      }
   }

   /*** Set font dimensions ***/

   SetField(font, FID_Bitmap, FT_POINTER, Bitmap);

   if (Self->Flags & TXF_WORDWRAP) {
      SetFields(font, FID_WrapWidth|TLONG, Self->AreaWidth - Self->LeftMargin - Self->RightMargin, TAGEND);
      font->Flags |= FTF_WORDWRAP;
   }

   valign = ((font->LineSpacing - font->MaxHeight) / 2);

   font->XCoord = Self->AreaX + Self->LeftMargin + Self->XPosition;
   font->YCoord = Self->AreaY + Self->TopMargin  + Self->YPosition + valign;
   font->AlignWidth  = Self->AreaWidth  - Self->LeftMargin - Self->RightMargin;
   font->AlignHeight = Self->AreaHeight - Self->TopMargin  - Self->BottomMargin;
   font->Flags |= FTF_DIRECTXY;

   GetField(font, FID_YOffset, FT_LONG, &alignoffset);

   /*** Set clipping area to match the text object ***/

   clipleft   = Bitmap->ClipLeft;
   clipright  = Bitmap->ClipRight;
   cliptop    = Bitmap->ClipTop;
   clipbottom = Bitmap->ClipBottom;
   if (Self->AreaX > Bitmap->ClipLeft) Bitmap->ClipLeft = Self->AreaX;
   if (Self->AreaY > Bitmap->ClipTop)  Bitmap->ClipTop  = Self->AreaY;
   if (Self->AreaX + Self->AreaWidth < Bitmap->ClipRight) Bitmap->ClipRight = Self->AreaX + Self->AreaWidth;
   if (Self->AreaY + Self->AreaHeight < Bitmap->ClipBottom) Bitmap->ClipBottom = Self->AreaY + Self->AreaHeight;

   /*** Clear the background if requested ***/

   if (Self->Flags & TXF_BACKGROUND) {
      ActionTags(MT_DrawRGBRectangle, Bitmap, Self->AreaX, Self->AreaY, Self->AreaWidth, Self->AreaHeight, Self->Background.Red,
         Self->Background.Green, Self->Background.Blue, TRUE);
   }

   /* Determine the start of the font area so that we can
   ** align the highlighting and the cursor correctly.
   */

   sx = Self->AreaX + Self->LeftMargin;

   /*** If an area has been selected, highlight it ***/

   selectrow    = -1;
   selectcolumn = -1;
   endrow       = -1;
   endcolumn    = -1;

   if (Self->Flags & TXF_AREASELECTED) {
      if ((Self->SelectRow != Self->CursorRow) OR (Self->SelectColumn != Self->CursorColumn)) {
         GetSelectedArea(Self, &selectrow, &selectcolumn, &endrow, &endcolumn);
      }
   }

   if ((Self->Effect IS EFX_TYPIST) AND (Self->EffectActive IS 1)) {
      /*** Typist effect support ***/

      chars = Self->CharCount;
      for (row=0; (row < Self->AmtLines) AND (font->YCoord < Bitmap->ClipBottom) AND (chars > 0); row++) {
         if (Self->Array[row].Length < chars) len = Self->Array[row].Length;
         else len = chars;

         chars -= len;

         if ((len > 0) AND (Self->Array[row].String)) {
            UBYTE buffer[len+1];
            for (i=0; i < len; i++) buffer[i] = Self->Array[row].String[i];
            buffer[i] = 0;

            SetField(font, FID_String, FT_STRING, buffer);
            Action(AC_Draw, font, NULL);
         }

         font->YCoord += font->LineSpacing;
      }
   }
   else {
      /*** Skip lines that are outside of the viewable area ***/

      if (Self->Flags & TXF_WORDWRAP) row = 0;
      else {
         for (row=0; (font->YCoord - valign + font->LineSpacing) <= Bitmap->ClipTop; row++) {
            font->YCoord += font->LineSpacing;
         }
      }

      basergb.Red   = font->ColourRGB->Red;
      basergb.Green = font->ColourRGB->Green;
      basergb.Blue  = font->ColourRGB->Blue;

      /*** In EDIT mode, there must always be at least 1 line so that we can print the cursor ***/

      amtlines = Self->AmtLines;
      if ((amtlines < 1) AND (Self->Flags & TXF_EDIT)) amtlines = 1;

      for (; (row < amtlines) AND (font->YCoord - valign < Bitmap->ClipBottom); row++) {

         /*** Do style management if there are tags listed against this line ***/

         currentfont = Self->Font;
         if (Self->Array[row].Tags) {
            style = NULL;

            if (CheckTag(Self->Array[row].Tags, TF_BOLD, NULL) IS ERR_Okay) style |= FTF_BOLD;

            if (CheckTag(Self->Array[row].Tags, TF_ITALICS, NULL) IS ERR_Okay) style |= FTF_ITALIC;

            if (style IS FTF_BOLD) {
               if (!(currentfont = Self->BoldFont)) {
                  if (CreateObject(ID_FONT, NF_CHILD, &Self->BoldFont, NULL,
                        FID_Face|TSTRING,  font->Face,
                        FID_Point|TLONG,   font->Point,
                        FID_Flags|TLONG,   font->Flags,
                        FID_Style|TSTRING, "Bold",
                        TAGEND) IS ERR_Okay) {
                     currentfont = Self->BoldFont;
                  }
               }
            }
            else if (style IS FTF_ITALIC) {
               if (!(currentfont = Self->ItalicFont)) {
                  if (CreateObject(ID_FONT, NF_CHILD, &Self->ItalicFont, NULL,
                        FID_Face|TSTRING,  font->Face,
                        FID_Point|TLONG,   font->Point,
                        FID_Flags|TLONG,   font->Flags,
                        FID_Style|TSTRING, "Italic",
                        TAGEND) IS ERR_Okay) {
                     currentfont = Self->ItalicFont;
                  }
               }
            }
            else if (style IS (FTF_BOLD|FTF_ITALIC)) {
               if (!(currentfont = Self->BIFont)) {
                  if (CreateObject(ID_FONT, NF_CHILD, &Self->BIFont, NULL,
                        FID_Face|TSTRING,  font->Face,
                        FID_Point|TLONG,   font->Point,
                        FID_Flags|TLONG,   font->Flags,
                        FID_Style|TSTRING, "Bold Italic",
                        TAGEND) IS ERR_Okay) {
                     currentfont = Self->BIFont;
                  }
               }
            }
         }

         if (currentfont != font) {
            currentfont->XCoord = font->XCoord;
            currentfont->YCoord = font->YCoord;
            currentfont->Bitmap = font->Bitmap;
            currentfont->LineSpacing = font->LineSpacing;
            SetField(currentfont, FID_WrapWidth, FT_LONG, &font->WrapWidth);
         }

         currentfont->ColourRGB->Red   = basergb.Red;
         currentfont->ColourRGB->Green = basergb.Green;
         currentfont->ColourRGB->Blue  = basergb.Blue;

         if (CheckTag(Self->Array[row].Tags, TF_COLOUR, &str) IS ERR_Okay) {
            SetFields(currentfont, FID_Colour|TSTRING, str, TAGEND);
         }

         if (CheckTag(Self->Array[row].Tags, TF_UNDERLINE, NULL) IS ERR_Okay) currentfont->Flags |= FTF_UNDERLINE;

         /*** Set the font string ***/

         if (Self->Flags & TXF_SECRET) {
            UBYTE buffer[Self->Array[row].Length+1];
            for (i=0; i < Self->Array[row].Length; i++) buffer[i] = '*';
            buffer[i] = 0;
            SetField(currentfont, FID_String, FT_STRING, buffer);
         }
         else if (Self->Flags & TXF_VARIABLE) {
            UBYTE buffer[Self->Array[row].Length+100];
            for (i=0; i < Self->Array[row].Length; i++) buffer[i] = Self->Array[row].String[i];
            buffer[i] = 0;
            StrTranslate(buffer, Self->Array[row].Length+100, NULL);
            SetField(currentfont, FID_String, FT_STRING, buffer);
         }
         else SetField(currentfont, FID_String, FT_STRING, Self->Array[row].String);


         /*** Draw any highlighting on this line ***/

         if (Self->Flags & TXF_SECRET) {
            /*** Highlighting is not allowed in secret mode ***/
         }
         else if ((row >= selectrow) AND (row <= endrow)) {
            width = Self->AreaWidth;
            if (row IS selectrow) {
               /*** First row ***/
               x = ColumnCoordinate(Self, row, selectcolumn);
               if (Self->Array[row].Length > 0) {
                  if (row IS endrow) {
                     calc.String = Self->Array[row].String;
                     calc.Length = endcolumn;
                     if (Action(MT_CalcStringWidth, currentfont, &calc) IS ERR_Okay) {
                        width = calc.Width - x + sx + Self->XPosition;
                     }
                     else width = 3;
                  }
                  else width = Self->Array[row].PixelLength - x + sx + Self->XPosition;
               }
               else width = 3;
            }
            else if (row < endrow) {
               /*** Middle row ***/
               x = currentfont->XCoord;
               if (Self->Array[row].Length > 0) width = Self->Array[row].PixelLength;
               else width = 3;
            }
            else {
               /*** End row ***/
               x = currentfont->XCoord;
               if ((Self->Array[row].Length > 0) AND (endcolumn > 0)) {
                  x = sx + Self->XPosition;
                  calc.String = Self->Array[row].String;
                  calc.Length = endcolumn;
                  Action(MT_CalcStringWidth, currentfont, &calc);
                  width = calc.Width - x + sx + Self->XPosition;
               }
               else width = 3;
            }

            ActionTags(MT_DrawRGBRectangle, Bitmap, x, currentfont->YCoord - valign + alignoffset,
               width, currentfont->LineSpacing, Self->Highlight.Red, Self->Highlight.Green, Self->Highlight.Blue, TRUE);
         }

         /*** Draw the cursor if the object is in edit mode ***/

         if ((row IS Self->CursorRow) AND (Self->KeyboardChannel IS TRUE) AND ((Self->CursorFlash % CURSOR_RATE) < (CURSOR_RATE>>1)) AND
             (Self->Flags & TXF_EDIT) AND (!Self->NoCursor)) {

            x = Self->AreaX + Self->LeftMargin + Self->XPosition;

            if (currentfont->AlignFlags & ALIGN_HORIZONTAL) {
               GetFields(currentfont, FID_Width|TLONG, &width, TAGEND);
               x += (Self->AreaWidth - width) / 2;
            }

            if ((str = currentfont->String)) {
               i = UTF8Length(str);
               column = Self->CursorColumn;
               if (column >= i) column = i;

               if (column > 0) {
                  calc.String = str;
                  calc.Length = column;
                  if (Action(MT_CalcStringWidth, currentfont, &calc) IS ERR_Okay) {
                     x += calc.Width;
                  }
               }
            }

            ActionTags(MT_DrawRGBRectangle, Bitmap, x, currentfont->YCoord - valign + alignoffset,
               Self->CursorWidth, currentfont->LineSpacing,
               Self->CursorRGB.Red, Self->CursorRGB.Green, Self->CursorRGB.Blue, TRUE);
         }

         /*** Draw the font now ***/

         if ((Self->Array[row].Length < 1) OR (!Self->Array[row].String)) {
            font->YCoord += font->LineSpacing;
         }
         else if (Self->Flags & TXF_WORDWRAP) {
            GetField(currentfont, FID_LineCount, FT_LONG, &i);
            if ((currentfont->YCoord - valign + (currentfont->LineSpacing * i)) >= Bitmap->ClipTop) {
               Action(AC_Draw, currentfont, NULL);
            }
            font->YCoord += currentfont->LineSpacing * i;
         }
         else {
            Action(AC_Draw, currentfont, NULL);
            font->YCoord += currentfont->LineSpacing;
         }
      }

      font->ColourRGB->Red   = basergb.Red;
      font->ColourRGB->Green = basergb.Green;
      font->ColourRGB->Blue  = basergb.Blue;
   }

   Bitmap->ClipLeft   = clipleft;
   Bitmap->ClipRight  = clipright;
   Bitmap->ClipTop    = cliptop;
   Bitmap->ClipBottom = clipbottom;
}

/*****************************************************************************

###ACTION###
Name:  Activate
Short: Opens the current text selection.

###DESCRIPTION###
<p>Calling the Activate action on a text object will cause it to send Activate
messages to the child objects that belong to the text object.</p>

###END###

*****************************************************************************/

static ERROR TEXT_Activate(struct Text *Self, APTR Void)
{
   LONG count, i;

   DPrintF("~Activate()","[Text:%d]", Self->Head.UniqueID);

   if ((count = TotalChildren(Self->Head.UniqueID))) {
      struct ChildEntry list[count];
      if (ListChildren(Self->Head.UniqueID, list, &count) IS ERR_Okay) {
         for (i=0; i < count; i++) {
            ActionMsg(AC_Activate, list[i].ObjectID, NULL);
         }
      }
   }

   StepBack();
   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:  AddLine()
Short: Adds a new line to any row position in a text object.

###DESCRIPTION###
<p>Lines can be added or inserted into a text object by using the AddLine()
method.  You need to provide the text string that you wish to use,
the line number that the text will be inserted into, and the length of the
text string.</p>

<p>If you set the Text argument to NULL, then an empty string will be
inserted into the line number.  If the Line argument is less than zero,
then the string will be added to the end of the Text.  If the Length is
set to -1, then the length will be calculated by counting the amount of
characters in the Text argument.</p>

<p>If the new line is visible within the text object's associated drawable,
that region of the drawable will be redrawn so that the new line is
displayed.</p>

###INPUT###
STRING/String: The text that you want to add.
LONG/Line: The number of the line at which the text should be inserted.
LONG/Length: The length of the Text string in bytes.

###RESULT###
ERR_Okay:        Success.
ERR_Args:        Invalid arguments were specified.
ERR_AllocMemory: The memory required to add the text string to the list was
                 unavailable.

###END###

*****************************************************************************/

static ERROR TEXT_AddLine(struct Text *Self, struct mtxAddLine *Args)
{
   if (!Args) return(ObjectError(Self, MTX_AddLine, ERR_Args));

   return(AddLine(Self, Args->String, Args->Line, Args->Length, FALSE));
}

/*****************************************************************************

###ACTION###
Name:  Clear
Short: Clears all content from the object.

###DESCRIPTION###
<p>You can delete all of the text information from a text object by calling
the Clear action.  All of the text data will be deleted from the object and
the graphics will be automatically updated as a result of calling this action.</p>

###END###

*****************************************************************************/

static ERROR TEXT_Clear(struct Text *Self, APTR Void)
{
   LONG i;

   if (Self->Array) {
      for (i=0; i < Self->AmtLines; i++) {
         if (Self->Array[i].String) FreeMemory(Self->Array[i].String);
         if (Self->Array[i].Tags)   FreeMemory(Self->Array[i].Tags);
      }
      FreeMemory(Self->Array);
      Self->Array = NULL;
   }

   /*** Reallocate the line array ***/

   Self->MaxLines = 50;
   if (AllocMemory(sizeof(struct TextLine) * Self->MaxLines, MEM_DATA, (APTR *)&Self->Array, NULL) != ERR_Okay) {
      return(ObjectError(Self, ERH_Clear, ERR_AllocMemory));
   }

   Self->AmtLines       = 0;
   Self->CursorRow      = 0;
   Self->CursorColumn   = 0;
   Self->YPosition      = 0;
   Self->XPosition      = 0;
   Self->ClickHeld      = FALSE;
   Self->SelectRow      = 0;
   Self->SelectColumn   = 0;
   Self->Flags         &= ~TXF_AREASELECTED;

   if (!Self->NoUpdate) {
      Redraw(Self);
      CalculateHScroll(Self);
      CalculateVScroll(Self);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Clipboard
Short: Full support for clipboard activity is provided through this action.
###END###

*****************************************************************************/

static ERROR TEXT_Clipboard(struct Text *Self, struct acClipboard *Args)
{
   struct acRead read;
   OBJECTPTR clipboard, file;
   STRING buffer;
   LONG size, i, row, column, endrow, endcolumn, pos, start;

   if ((!Args) OR (!Args->ClipboardID) OR (!Args->Mode)) {
      return(ObjectError(Self, ERH_Clipboard, ERR_Args));
   }

   if ((Args->Mode IS CLIP_CUT) OR (Args->Mode IS CLIP_COPY)) {

      if (Args->Mode IS CLIP_CUT) DPrintF("~Clipboard:","[Text:%d] Operation: Cut", Self->Head.UniqueID);
      else DPrintF("~Clipboard:","[Text:%d] Operation: Copy", Self->Head.UniqueID);

      /*** Calculate the length of the highlighted text ***/

      if ((Self->Flags & TXF_AREASELECTED) AND ((Self->SelectRow != Self->CursorRow) OR
          (Self->SelectColumn != Self->CursorColumn))) {

         GetSelectedArea(Self, &row, &column, &endrow, &endcolumn);
         column = UTF8CharacterOffset(Self->Array[row].String, column);
         endcolumn = UTF8CharacterOffset(Self->Array[endrow].String, endcolumn);

         size = 0;
         for (i=row; i <= endrow; i++) size += Self->Array[i].Length + 1;
         if (AllocMemory(size+1, MEM_STRING, &buffer, NULL) IS ERR_Okay) {
            pos = 0;
            start = row;

            /*** Copy the selected area into the buffer ***/

            if (Self->Array[row].String) {
               if (row IS endrow) {
                  while (column < endcolumn) buffer[pos++] = Self->Array[row].String[column++];
               }
               else while (column < Self->Array[row].Length) buffer[pos++] = Self->Array[row].String[column++];
            }

            if (++row <= endrow) {
               for (; row < endrow; row++) {
                  buffer[pos++] = '\n';
                  if (Self->Array[row].String) {
                     for (i=0; i < Self->Array[row].Length; i++) buffer[pos++] = Self->Array[row].String[i];
                  }
               }
               buffer[pos++] = '\n';
               if (Self->Array[row].String) {
                  for (i=0; i < endcolumn; i++) buffer[pos++] = Self->Array[row].String[i];
               }
            }

            buffer[pos] = 0;

            /*** Send the text to the clipboard object ***/

            if (AccessObject(Args->ClipboardID, 5000, &clipboard) IS ERR_Okay) {
               ActionTags(MT_AddClipboardText, clipboard, buffer);
               ReleaseObject(clipboard);
            }

            FreeMemory(buffer);

            /*** Delete the highlighted text if the CUT mode was used ***/

            if (Args->Mode IS CLIP_CUT) DeleteSelectedArea(Self);
            else {
               //Self->Flags &= ~TXF_AREASELECTED;
               //DrawLines(Self, start, endrow - start + 1);
            }
         }
         else {
            ObjectError(Self, ERH_Clipboard, ERR_AllocMemory);
            StepBack();
            return(ERR_AllocMemory);
         }
      }

      StepBack();
      return(ERR_Okay);
   }
   else if (Args->Mode IS CLIP_PASTE) {
      DPrintF("~Clipboard:","[Text:%d] Operation: Paste");

      if (!(Self->Flags & TXF_EDIT)) {
         DPrintF("@Clipboard:","[Text:%d] Edit mode is not enabled, paste operation aborted.", Self->Head.UniqueID);
         return(ERR_Failed);
      }

      if (AccessObject(Args->ClipboardID, 5000, &clipboard) IS ERR_Okay) {
      	 struct mtGetClipboardFiles get;

      	 get.Datatype = CLIP_TEXT;
         get.Index = NULL;
         if (Action(MT_GetClipboardFiles, clipboard, &get) IS ERR_Okay) {
            if (CreateObject(ID_FILE, NULL, &file, NULL,
               FID_Location|TSTRING, get.Files[0],
               FID_Flags|TLONG,      FL_READ,
               TAGEND) IS ERR_Okay) {

               if ((GetField(file, FID_Size, FT_LONG, &size) IS ERR_Okay) AND (size > 0)) {
                  if (AllocMemory(size, MEM_STRING, &buffer, NULL) IS ERR_Okay) {
                     read.Buffer = buffer;
                     read.Length = size;
                     if (Action(AC_Read, file, &read) IS ERR_Okay) {
                        ActionTags(AC_DataChannel, Self, GetUniqueID(Self), DATA_TEXT, 1,
                           buffer, read.Result, 1);
                     }
                     else ObjectError(Self, ERH_Clipboard, ERR_Read);
                     
                     FreeMemory(buffer);
                  }
                  else ObjectError(Self, ERH_Clipboard, ERR_AllocMemory);
               }
               else DPrintF("@Clipboard:","[Text:%d] The clipboard file is empty.", Self->Head.UniqueID);

               acFree(file);
            }
         }

         ReleaseObject(clipboard);
         StepBack();
         return(ERR_Okay);
      }
      else {
         ObjectError(Self, ERH_Clipboard, ERR_AccessObject);
         StepBack();
         return(ERR_AccessObject);
      }
   }
   else return(ObjectError(Self, ERH_Clipboard, ERR_Args));
}

/*****************************************************************************
** Text: ClosingTag
*/

static ERROR TEXT_ClosingTag(struct Text *Self, APTR Void)
{
   Self->Tag--;
   Self->NoUpdate--;

   CalculateHScroll(Self);
   CalculateVScroll(Self);

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  DataChannel
Short: Text data can be sent to a text object via data channels.

###DESCRIPTION###
<p>One of the easiest ways to send text information to a text object is to
do so via data channels.  The Text class currently supports the DATA_TEXT
and DATA_XML types for this purpose.  If the text contains return codes, the
data will be split into multiple lines.</p>

<p>The drawable that is associated with the Text object will be redrawn
as a result of calling this action.</p>

###RESULT###
ERR_Okay:        Success.
ERR_Args:        Invalid arguments were specified.
ERR_AllocMemory: The Text's memory buffer could not be expanded.
ERR_Mismatch:    The data type that was passed to the action is not supported by the Text class.

###END###

*****************************************************************************/

static ERROR TEXT_DataChannel(struct Text *Self, struct acDataChannel *Args)
{
   STRING line, str, buffer;
   LONG i, j, len, linestart, pos, end, size;

   if (!Args) return(ObjectError(Self, ERH_DataChannel, ERR_Args));

   if ((Args->DataType IS DATA_KEYPRESS) AND (Args->Version IS 1)) {
      for (j=0; j < Args->TotalEntries; j++) {
         if (((struct dcKeyEntry *)Args->Buffer)[j].Flags & KQ_PRESSED) {
            EditKeypress(Self, ((struct dcKeyEntry *)Args->Buffer)[j].Flags, ((struct dcKeyEntry *)Args->Buffer)[j].Value);
         }
      }
      return(ERR_Okay);
   }
   else if ((Args->DataType IS DATA_TEXT) AND (Args->Version <= 1)) {
      if (!Args->Size) return(ERR_Okay);

      DPrintF("~6DataChannel()","[Text:%d] Inserting text data of size %d.", Self->Head.UniqueID, Args->Size);

      Self->NoUpdate++;

      linestart = Self->CursorRow;

      line = Args->Buffer;

      if ((Self->Flags & TXF_EDIT) AND (Self->Flags & TXF_AREASELECTED)) {
         DeleteSelectedArea(Self);
      }

      if ((Self->Flags & TXF_EDIT) AND (Self->AmtLines > 0)) {

         if ((linestart = Self->CursorRow) < 0) linestart = 0;
         for (len=0; (line[len] != '\n') AND (line[len] != '\r') AND (len < Args->Size); len++); /* Length of the first line */

         if (AllocMemory(Self->Array[Self->CursorRow].Length + len + 1, MEM_STRING|MEM_NOCLEAR, &str, NULL) IS ERR_Okay) {

            if ((len >= Args->Size) OR (Self->LineLimit IS 1)) {
               j = 0;
               for (i=0; (i < Self->CursorColumn) AND (i < Self->Array[Self->CursorRow].Length); i++) str[i] = Self->Array[Self->CursorRow].String[j++];
               for (pos=0; pos < len; pos++) str[i++] = line[pos];
               Self->CursorColumn = i;
               while (j < Self->Array[Self->CursorRow].Length) str[i++] = Self->Array[Self->CursorRow].String[j++];
               ActionTags(MTX_ReplaceLine, Self, str, Self->CursorRow, i);
            }
            else {
               /*** Replace the first line ***/

               j = 0;
               for (i=0; (i < Self->CursorColumn) AND (i < Self->Array[Self->CursorRow].Length); i++) str[i] = Self->Array[Self->CursorRow].String[j++];
               for (pos=0; pos < len; pos++) str[i++] = line[pos];
               end = i;
               while (j < Self->Array[Self->CursorRow].Length) str[i++] = Self->Array[Self->CursorRow].String[j++];
               str[i] = 0;
               ActionTags(MTX_ReplaceLine, Self, str, Self->CursorRow, Self->CursorColumn + len);

               /*** Add further lines ***/

               Self->CursorRow++;
               pos++;
               if ((pos < Args->Size) AND (line[pos] IS '\r')) pos++;

               if (pos < Args->Size) {
                  while (pos < Args->Size) {
                     for (len=0; (line[pos+len] != '\n') AND (line[pos+len] != '\r') AND (pos+len < Args->Size); len++);
                     ActionTags(MTX_AddLine, Self, line+pos, Self->CursorRow++, len);

                     if ((pos+len < Args->Size) AND (line[pos+len] IS '\r')) len++;

                     if (pos+len < Args->Size) {
                        len++;
                        if (pos+len >= Args->Size) ActionTags(MTX_AddLine, Self, NULL, Self->CursorRow++, NULL); /* Blank line */
                     }
                     pos += len;
                  }
               }
               else ActionTags(MTX_AddLine, Self, NULL, Self->CursorRow++, NULL); /* Blank line */

               Self->CursorRow--;

               /*** Replace the last line ***/

               for (len=0; str[end+len]; len++);
               if (Self->Array[Self->CursorRow].Length + len) {
                  if (AllocMemory(Self->Array[Self->CursorRow].Length + len, MEM_STRING|MEM_NOCLEAR, &buffer, NULL) IS ERR_Okay) {
                     for (i=0; i < Self->Array[Self->CursorRow].Length; i++) buffer[i] = Self->Array[Self->CursorRow].String[i];
                     Self->CursorColumn = i;
                     for (j=0; j < len; j++) buffer[i++] = str[end+j];
                     ActionTags(MTX_ReplaceLine, Self, buffer, Self->CursorRow, i);
                     FreeMemory(buffer);
                  }
               }
            }

            FreeMemory(str);
         }
      }
      else {
         STRING str;
         WORD linefeed, trailing_line;

         if ((linestart = Self->AmtLines - 1) < 1) linestart = 0;

         trailing_line = FALSE;
         for (pos=0; pos < Args->Size; ) {
            /*** If we have run out of lines, expand the line list ***/

            line = Args->Buffer + pos;
            size = Args->Size - pos;
            linefeed = 1; /* Normally a linefeed will consist of just the return character */

            if (Self->Tag) {
               /* NOTE: When text is encapsulated inside <text>...</text> tags, ALL whitespace
               ** is converted into spaces in this processing routine.  If the user wishes to
               ** force return codes, he should use the \n character to do this, or a double-return
               ** if WORDWRAP is enabled.
               */
               for (len=0; len < size; len++) {
                  if (Self->Flags & TXF_WORDWRAP) {
                     /*** Two returns indicate a line break when wordwrap is enabled ***/
                     if ((line[len] IS '\n') AND (len < size-1) AND (line[len+1] IS '\n')) break;
                  }

                  if ((line[len] IS '\\') AND (len+1 < size) AND (line[len+1] IS 'n')) {
                     linefeed = 2;
                     break;
                  }
               }
            }
            else {
               for (len=0; (line[len] != '\n') AND (len < size); len++);
               /* Check for a trailing line (a return code at the
               ** very end of the data feed)
               */
               if ((len IS size-1) AND (line[len] IS '\n')) trailing_line = TRUE;
            }

            if (len > 0) {
               if (AllocMemory(len+1, MEM_STRING|MEM_NOCLEAR, &str, NULL) IS ERR_Okay) {
                  j = 0;
                  if (Self->Tag) {
                     for (i=0; i < len; i++) {
                        if ((line[i] IS '\\') AND (i+1 < len)) {
                           if (line[i+1] IS '\\') { str[j++] = '\\'; i++; }
                           else if (line[i+1] IS 'n') i++; /* Skip "\n" character strings */
                        }
                        else if (line[i] IS '\r'); /* Ignore carriage returns */
                        else if (line[i] IS '\t') str[j++] = '\t'; /* Accept tabs, don't convert to ' ' */
                        else if (line[i] <= 0x20) {
                           if ((j > 0) AND (str[j-1] IS ' ')); /* Do nothing if the last character was a space */
                           else str[j++] = ' ';  /* Turn all other whitespace into spaces */
                        }
                        else str[j++] = line[i]; /* Accept standard character */
                     }
                  }
                  else for (i=0; i < len; i++) {
                     if (line[i] IS '\r') continue;
                     str[j++] = line[i];
                  }
                  str[j] = 0;

                  AddLine(Self, str, -1, j, TRUE);
               }
            }
            else {
               /*** Add a blank line ***/
               AddLine(Self, "", -1, 0, FALSE);
            }

            if (trailing_line) {
               AddLine(Self, "", -1, 0, FALSE);
               break;
            }

            pos += len + linefeed;
         }
      }

      Self->NoUpdate--;

      if (!Self->NoUpdate) {
         CalculateHScroll(Self);
         CalculateVScroll(Self);

         DrawLines(Self, linestart, Self->AmtLines - linestart);
         ViewCursor(Self);
      }

      StepBack();
      return(ERR_Okay);
   }
   else if ((Args->DataType IS DATA_XML) AND (Args->Version <= 1)) {
      struct XMLTag **xmllist, *tag;
      LONG linestart, itemcount;

      DPrintF("5DataChannel()","[Text:%d] Received an XML statement of %d bytes.", Self->Head.UniqueID, Args->Size);

      /* Accepted XML tags are:
      **
      ** ITEM: Lines can be grouped under separate item tags.
      ** LINK: If a line should have a special program or script setting to
      **       override default settings, this can be done through a link.
      **       Attributes: Program, ProgramArgs, Script, ScriptTarget.
      ** B:    Bold text.
      ** I:    Italic text.
      ** FONT: Indicates a change in the font that overrides default settings.
      **       Attributes: Colour
      **
      ** Please note that XML tags currently affect the entire line - a
      ** tag cannot affect a selected portion of text :-/
      */

      if ((linestart = Self->AmtLines - 1) < 1) linestart = 0;  //linestart = Self->AmtLines;

      if (!Self->XML) {
         if (CreateObject(ID_XML, NF_CHILD, &Self->XML, NULL,
            FID_Statement|TSTRING, Args->Buffer,
            NULL) != ERR_Okay) {
            return(ObjectError(Self, ERH_DataChannel, ERR_CreateObject));
         }
      }
      else if (SetField(Self->XML, FID_Statement, FT_POINTER, Args->Buffer) != ERR_Okay) {
         return(ERR_SetField);
      }

      if ((GetField(Self->XML, FID_Tags, FT_POINTER, &xmllist) IS ERR_Okay) AND (xmllist)) {

         /*** Search for <item> tags and add them as individual lines ***/

         itemcount = 0;
         for (tag=xmllist[0]; tag; tag=tag->Next) {
            if (StrCompare("item", tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
               AddXML(Self, tag, FALSE, -1);
               itemcount++;
            }
            else if (StrCompare("pre", tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
               AddXML(Self, tag, TRUE, -1);
               itemcount++;
            }
         }

         /* If there were no <item> tags in the XML statement,
         ** assume that the XML statement counts as one single item.
         */

         if (!itemcount) AddXML(Self, xmllist[0], FALSE, -1);
      }

      return(ERR_Okay);
   }
   else {
      DPrintF("DataChannel()","[Text:%d] DataID %d / Version %d not supported.", GetUniqueID(Self), Args->DataType, Args->Version);
      return(ERR_Mismatch);
   }
}

/*****************************************************************************

###METHOD###
Name:  DeleteLine()
Short: Deletes any line number.

###DESCRIPTION###
<p>This method deletes lines from a text object.  You only need to specify
the line number to have it deleted.  If the line number does not exist, then
the call will fail.  The text graphic will be updated as a result of calling
this method.</p>

###INPUT###
LONG/Line: The line number that you want to delete.  If negative, the last line will be deleted.

###RESULT###
ERR_Okay: The line was deleted.
ERR_Args: The Line value was out of the valid range.

###END###

*****************************************************************************/

static ERROR TEXT_DeleteLine(struct Text *Self, struct mtxDeleteLine *Args)
{
   if (Self->AmtLines < 1) return(ERR_Okay);

   if (Self->CursorRow IS Args->Line) MoveCursor(Self, Self->CursorRow, 0);

   if ((!Args) OR (Args->Line < 0)) {
      /*** Delete the line at the very end of the list ***/
      if (Self->Array[Self->AmtLines-1].String) FreeMemory(Self->Array[Self->AmtLines-1].String);
      if (Self->Array[Self->AmtLines-1].Tags) FreeMemory(Self->Array[Self->AmtLines-1].Tags);
      ClearMemory(Self->Array + Self->AmtLines - 1, sizeof(Self->Array[0]));

      if (Self->Flags & TXF_AREASELECTED) {
         Self->Flags &= ~TXF_AREASELECTED;
         Redraw(Self);
      }
      else RedrawLine(Self, Self->AmtLines-1);

      Self->AmtLines--;
      if (Self->CursorRow >= Self->AmtLines) MoveCursor(Self, Self->AmtLines-1, Self->CursorColumn);
   }
   else {
      if (Args->Line >= Self->AmtLines) return(ObjectError(Self, MTX_DeleteLine, ERR_Args));

      if (Self->Array[Args->Line].String) FreeMemory(Self->Array[Args->Line].String);
      if (Self->Array[Args->Line].Tags) FreeMemory(Self->Array[Args->Line].Tags);
      ClearMemory(Self->Array + Args->Line, sizeof(Self->Array[0]));

      if (Args->Line < Self->AmtLines - 1) {
         CopyMemory(Self->Array + Args->Line + 1, Self->Array + Args->Line, sizeof(Self->Array[0]) * (Self->AmtLines - Args->Line - 1));
      }

      Self->AmtLines--;

      if (Self->CursorRow >= Self->AmtLines) MoveCursor(Self, Self->AmtLines-1, Self->CursorColumn);

      if (Self->Flags & TXF_AREASELECTED) {
         Self->Flags &= ~TXF_AREASELECTED;
         Redraw(Self);
      }
      else DrawLines(Self, Args->Line, Self->AmtLines - Args->Line + 1);
   }

   CalculateHScroll(Self);
   CalculateVScroll(Self);
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Disable
Short: Disables object functionality.
###END###

*****************************************************************************/

static ERROR TEXT_Disable(struct Text *Self, APTR Void)
{
   Self->Flags |= TXF_DISABLED;
   return(ERR_Okay);
}

/*****************************************************************************
** Text: Draw
*/

static ERROR TEXT_Draw(struct Text *Self, APTR Void)
{
   if (Self->DrawableID) {
      Redraw(Self);
      return(ERR_Okay);
   }
   else return(ERR_FieldNotSet);
}

/*****************************************************************************

###ACTION###
Name:  Enable
Short: Enables object functionality.
###END###

*****************************************************************************/

static ERROR TEXT_Enable(struct Text *Self, APTR Void)
{
   Self->Flags &= ~TXF_DISABLED;
   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:  FindText
Short: Searches for keywords.

###DESCRIPTION###
<p>This method is used to search for keywords within text objects.  It is
possible to call this function once to find the first instance of the keyword,
or repeatedly if multiple keywords need to be found.</p>

<p>A large number of optional flags are available to refine the type of search
that is performed, as listed in the following table:</p>

<table border="1" bordercolordark="#000000" cellspacing="0" cellpadding="2" bordercolorlight="#000000" width="80%" align="center">
<tr><td class="tableheader" bgcolor="#eeeeee">Flag</td><td class="tableheader" bgcolor="#eeeeee">Description</td></tr>
<tr><td width="10%">FTF_BACKWARDS</td><td>Start the search from the end of the text and work back to the beginning.</td></tr>
<tr><td width="10%">FTF_CASE</td><td>Use case sensitivity when looking for the keyword.</td></tr>
<tr><td width="10%">FTF_FROMCURSOR</td><td>Start the search from the current cursor position.</td></tr>
<tr><td width="10%">FTF_HIGHLIGHT</td><td>This flag is used to highlight the keyword when it is found.</td></tr>
<tr><td width="10%">FTF_MOVECURSOR</td><td>Move the cursor to the end of the keyword position when it is found.</td></tr>
<tr><td width="10%">FTF_SELECTION</td><td>Search only within the current text selection.  If no text has been selected, do not perform the search.</td></tr>
<tr><td width="10%">FTF_RESTART</td><td>Restart the search from the beginning of the text data if the keyword is not found from the current position.</td></tr>
</table>

<p>If the keyword is found, the Row and Column parameters will be updated
to reflect the location of the keyword and ERR_Okay will be returned.  In all
other cases, the Row and Column parameters will be driven to zero and
ERR_Search will be returned to indicate that the keyword was not found.</p>

###INPUT###
STRING/String: The keyword that you want to search for.
LONG/Flags:    Optional flags.
LONG/Row:      This parameter will be updated to reflect the row at which the keyword was found.
LONG/Column:   This parameter will be updated to reflect the column at which the keyword was found.

###RESULT###
ERR_Okay:   The keyword was found.
ERR_Args:   Invalid arguments were specified.
ERR_Search: The keyword was not found.

###END###

*****************************************************************************/

static ERROR TEXT_FindText(struct Text *Self, struct mtxFindText *Args)
{
   STRING str;
   LONG start_row, row, column, endrow, endcolumn, pos, i;
   BYTE cont;

   DPrintF("FindText()","[Text:%d] Searching for \"%.30s\".", Self->Head.UniqueID, Args->String);

   if ((!Args) OR (!Args->String) OR (!*Args->String)) {
      return(ObjectError(Self, MTX_FindText, ERR_Args));
   }

   Args->Row = 0;
   Args->Column = 0;
   if (Args->Flags & FTF_RESTART) cont = TRUE;
   else cont = FALSE;

   if (Self->AmtLines < 1) return(ERR_Search);

   if (Args->Flags & FTF_SELECTION) {
      /*** Search the selected area ***/
      if (Self->Flags & TXF_AREASELECTED) {
         GetSelectedArea(Self, &row, &column, &endrow, &endcolumn);
      }
      else return(ERR_Search);
   }
   else if (Args->Flags & FTF_FROMCURSOR) {
      row    = Self->CursorRow;
      column = Self->CursorColumn;
      endrow = Self->AmtLines - 1;
      endcolumn = Self->Array[endrow].Length;
   }
   else {
      row       = 0;
      column    = 0;
      endrow    = Self->AmtLines - 1;
      endcolumn = Self->Array[endrow].Length;
   }

   if ((row >= Self->AmtLines) OR (row < 0)) return(ERR_Search);
   if (column > Self->Array[row].Length) column = Self->Array[row].Length;

   if (Args->Flags & FTF_BACKWARDS) {

   }
   else {
restart:
      start_row = row;

      while (row <= endrow) {
         /*** Search the string ***/

         str = Self->Array[row].String;
         if (Args->Flags & FTF_CASE) {
            for (pos=column; pos < Self->Array[row].Length; pos++) {
               for (i=0; Args->String[i]; i++) {
                  if (str[pos+i] != Args->String[i]) break;
               }
               if (!Args->String[i]) break; /* Keyword found */
            }
         }
         else for (pos=column; pos < Self->Array[row].Length; pos++) {
            for (i=0; Args->String[i]; i++) {
               if (UCase(str[pos+i]) != UCase(Args->String[i])) break;
            }
            if (!Args->String[i]) break; /* Keyword found */
         }

         if (pos < Self->Array[row].Length) {
            if ((row < endrow) OR ((row IS endrow) AND (pos < endcolumn))) {
               Args->Row    = row;
               Args->Column = pos;

               if (Args->Flags & FTF_HIGHLIGHT) {
                  /*** Highlight the text when we find it ***/

                  if (Self->Flags & TXF_AREASELECTED) {
                     /*** Remove the current highlighted area ***/
                     Self->Flags &= ~TXF_AREASELECTED;
                     Self->NoCursor++;
                     DrawLines(Self, Self->SelectRow, Self->CursorRow - Self->SelectRow + 1);
                     Self->NoCursor--;
                  }
                  else RemoveCursor(Self);

                  Self->Flags |= TXF_AREASELECTED;
                  Self->SelectRow    = Args->Row;
                  Self->SelectColumn = Args->Column;
                  Self->CursorRow    = Self->SelectRow;
                  Self->CursorColumn = Self->SelectColumn + StrLength(Args->String);

                  DrawLines(Self, Self->SelectRow, Self->CursorRow - Self->SelectRow + 1);

                  ViewSelection(Self);
               }
               else if (Args->Flags & FTF_MOVECURSOR) {
                  /*** Move the cursor to the text position ***/
                  MoveCursor(Self, Args->Row, Args->Column);
               }

               return(ERR_Okay);
            }
         }
         column = 0;
         row++;
      }

      if (cont IS TRUE) {
         cont = FALSE;
         row = 0;
         endrow = start_row;
         goto restart;
      }
   }

   return(ERR_Search);
}

/*****************************************************************************

###ACTION###
Name:  Focus
Short: Calling this action will activate keyboard input.
###END###

*****************************************************************************/

static ERROR TEXT_Focus(struct Text *Self, APTR Void)
{
   return(ActionMsg(AC_Focus, Self->DrawableID, NULL));
}

/*****************************************************************************
** Text: Free
*/

static ERROR TEXT_Free(struct Text *Self, APTR Void)
{
   OBJECTPTR object;
   LONG i;

   Self->Visible = FALSE;

   UnsubscribeTimer(Self->Head.UniqueID);

   if ((Self->InputID) AND (Self->KeyboardChannel IS TRUE)) {
      if (AccessObject(Self->InputID, 5000, &object) IS ERR_Okay) {
         UnsubscribeChannel(object, Self->Head.UniqueID);
         ReleaseObject(object);
      }
      Self->InputID = NULL;
   }

   if ((Self->FocusID) AND (Self->FocusID != Self->DrawableID)) {
      if (AccessObject(Self->FocusID, 5000, &object) IS ERR_Okay) {
         UnsubscribeAction(object, NULL, GetUniqueID(Self));
         ReleaseObject(object);
      }
   }

   if (Self->DrawableID) {
      if (AccessObject(Self->DrawableID, 5000, &object) IS ERR_Okay) {
         ActionTags(MT_SetRenderCallback, object, Self, NULL);
         UnsubscribeAction(object, NULL, GetUniqueID(Self));
         ReleaseObject(object);
      }
   }

   if (Self->PointerLocked) {
      RestoreCursor(PTR_DEFAULT, Self->Head.UniqueID);
      Self->PointerLocked = FALSE;
   }

   if (Self->Array) {
      for (i=0; i < Self->AmtLines; i++) {
         if (Self->Array[i].String) FreeMemory(Self->Array[i].String);
         if (Self->Array[i].Tags) FreeMemory(Self->Array[i].Tags);
      }
      FreeMemory(Self->Array);
      Self->Array = NULL;
   }
   Self->AmtLines = 0;
   Self->MaxLines = 0;

   if (Self->Frames) {
      FreeMemory(Self->Frames);
      Self->Frames = NULL;
   }

   if (Self->Location) {
      FreeMemory(Self->Location);
      Self->Location = NULL;
   }

   if (Self->History) {
      FreeMemory(Self->History);
      Self->History = NULL;
   }

   if (Self->XML)        { Action(AC_Free, Self->XML, NULL);        Self->XML = NULL; }
   if (Self->Font)       { Action(AC_Free, Self->Font, NULL);       Self->Font = NULL; }
   if (Self->BoldFont)   { Action(AC_Free, Self->BoldFont, NULL);   Self->BoldFont = NULL; }
   if (Self->ItalicFont) { Action(AC_Free, Self->ItalicFont, NULL); Self->ItalicFont = NULL; }
   if (Self->BIFont)     { Action(AC_Free, Self->BIFont, NULL);     Self->BIFont = NULL; }

   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:  GetLine()
Short: Returns the string content of any given line.

###DESCRIPTION###
<p>This method can be used to get the string associated with any given line
number.  You may choose to provide a buffer space for the method to output the
string data to, or you may set the Buffer argument to NULL to have the
method allocate a memory block containing the string.  If you are providing
a buffer, make sure that the Length argument is set to the correct buffer
size.  In the case of allocated buffers, the Length argument will
be updated to reflect the length of the allocation (including the NULL
byte).</p>

###INPUT###
LONG/Line:     The line number that you want to retrieve.  Must be zero or greater.
STRING/Buffer: Point this argument to a buffer space for the string result, or
               set to NULL if a buffer should be allocated by the method.
LONG/Length: Set this argument to the length of the buffer that you have
             provided, or set to NULL if a buffer is to be allocated.

###RESULT###
ERR_Okay:       The method executed successfully.
ERR_Args:       Invalid arguments detected.
ERR_OutOfRange: The line number that you specified was outside of the valid
                range of line numbers.
ERR_AllocMemory: The necessary amount of buffer space could not be allocated.

###END###

*****************************************************************************/

static ERROR TEXT_GetLine(struct Text *Self, struct mtxGetLine *Args)
{
   if (!Args) return(ObjectError(Self, MTX_GetLine, ERR_Args));

   if (Args->Line >= Self->AmtLines) {
      DPrintF("@GetLine()","[Text:%d] Cannot retrieve line %d (%d lines available).", Args->Line, Self->AmtLines);
      return(ERR_OutOfRange);
   }

   if (!Args->Buffer) {
      Args->Length = Self->Array[Args->Line].Length + 1;
      if (AllocMemory(Args->Length, MEM_STRING|MEM_NOCLEAR|MEM_TASK, (APTR *)&Args->Buffer, NULL) != ERR_Okay) {
         return(ERR_AllocMemory);
      }
   }

   StrCopy(Self->Array[Args->Line].String, Args->Buffer, Args->Length);

   return(ERR_Okay);
}

/*****************************************************************************
** Text: Hide
*/

static ERROR TEXT_Hide(struct Text *Self, APTR Void)
{
   if (Self->Visible IS TRUE) {
      Self->Visible = FALSE;
      Redraw(Self);
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Text: Init
*/

static ERROR TEXT_Init(struct Text *Self, APTR Void)
{
   struct Variable value;
   struct acRead read;
   struct Render *drawable;
   OBJECTPTR file, object;
   LONG size, i;
   STRING line;

   if (!Self->DrawableID) return(ObjectError(Self, ERH_Init, ERR_UnsupportedOwner));

   if (!Self->FocusID) Self->FocusID = Self->DrawableID;

   /*** Initialise the Font ***/

   Self->Font->Flags |= FTF_DIRECTXY;
   if (Self->Flags & TXF_WORDWRAP) Self->Font->Flags |= FTF_WORDWRAP;
   if (Action(AC_Init, Self->Font, NULL) != ERR_Okay) {
      return(ObjectError(Self, ERH_Init, ERR_Init));
   }

   /*** Now that we have a font, we can calculate the pixel widths of each existing text line ***/

   for (i=0; i < Self->AmtLines; i++) {
      Self->Array[i].PixelLength = CalculateWidth(Self, Self->Array[i].String, Self->Array[i].Length);
   }

   /*** Load a text file into the line array if required ***/

   if (Self->Location) {
      if (CreateObject(ID_FILE, NF_CHILD, &file, NULL,
         FID_Flags|TLONG,      FL_READ,
         FID_Location|TSTRING, Self->Location,
         TAGEND) IS ERR_Okay) {

         if ((GetField(file, FID_Size, FT_LONG, &size) IS ERR_Okay) AND (size > 0)) {
            if (AllocMemory(size, MEM_STRING|MEM_NOCLEAR, (APTR *)&line, NULL) IS ERR_Okay) {
               read.Buffer = line;
               read.Length = size;
               if (Action(AC_Read, file, &read) IS ERR_Okay) {
                  Action(AC_Clear, Self, NULL);
                  ActionTags(AC_DataChannel, Self, Self->Head.UniqueID, DATA_TEXT,
                     1, line, read.Result, 1);
               }
               FreeMemory(line);
            }
            else {
               Action(AC_Free, file, NULL);
               return(ObjectError(Self, ERH_Init, ERR_AllocMemory));
            }
         }
         Action(AC_Free, file, NULL);
      }
   }

   /*** If coordinates are missing, set them to zero ***/

   value.Type = FD_DOUBLE;
   value.Double = 0;
   if ((!(Self->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X)))) {
      if (!((Self->Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET)) AND (Self->Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH)))) {
         SET_X(Self, &value);
      }
   }

   if ((!(Self->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y)))) {
      if (!((Self->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET)) AND (Self->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT)))) {
         SET_Y(Self, &value);
      }
   }

   if ((!(Self->Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) AND (!(Self->Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH)))) {
      SET_XOffset(Self, &value);
   }

   if ((!(Self->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) AND (!(Self->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT)))) {
      SET_YOffset(Self, &value);
   }

   /*** Allocate a history buffer if TXF_HISTORY is enabled ***/

   if (Self->Flags & TXF_HISTORY) {
      if (AllocMemory(Self->HistorySize * sizeof(struct TextHistory), MEM_DATA, &Self->History, NULL) != ERR_Okay) {
         Self->Flags &= ~TXF_HISTORY;
      }
   }

   /*** Setup keyboard ***/

   if (!Self->InputID) {
      if (FastFindObject("SystemKeyboard", NULL, &Self->InputID, 1, NULL) != ERR_Okay) {
         DPrintF("@Init:","[Text] Warning: Could not find any SystemKeyboard objects.");
      }
   }

   if (Self->Flags & TXF_GLOBALEDITING) {
      struct acActionNotify notify;

      DPrintF("Init:","[Text] Using global editing mode.");

      notify.ActionID = AC_Focus;
      notify.Error    = ERR_Okay;
      Action(AC_ActionNotify, Self, &notify);
   }
   else if (AccessObject(Self->FocusID, 5000, &object) IS ERR_Okay) {
      SubscribeActionTags(object, Self,
         AC_Focus,
         AC_LostFocus,
         TAGEND);

      /* Test the focus object to see if it already has the focus and
      ** if so, activate text input.
      */

      if (object->ClassID IS ID_RENDER) {
         if (((struct Render *)object)->Flags & RNF_HASFOCUS) {
            struct acActionNotify notify;
            notify.ActionID = AC_Focus;
            notify.Error    = ERR_Okay;
            Action(AC_ActionNotify, Self, &notify);
         }
      }

      ReleaseObject(object);
   }

   /*** Subscribe to the drawable ***/

   if (AccessObject(Self->DrawableID, 5000, (OBJECTPTR *)&drawable) IS ERR_Okay) {

      Self->DrawableWidth = drawable->Width;
      Self->DrawableHeight = drawable->Height;

      ActionTags(MT_SetRenderCallback, drawable, Self, &DrawText);

      SubscribeActionTags(drawable, Self,
         AC_Disable,
         AC_Enable,
         AC_Redimension,
         AC_UserClick,
         AC_UserClickRelease,
         AC_UserMovement,
         TAGEND);

      value.Type = FD_LARGE;
      GET_X(Self, &value); Self->AreaX = value.Large;
      GET_Y(Self, &value); Self->AreaY = value.Large;
      GET_Width(Self, &value);  Self->AreaWidth  = value.Large;
      GET_Height(Self, &value); Self->AreaHeight = value.Large;

      ReleaseObject(drawable);
   }
   else return(ObjectError(Self, ERH_Init, ERR_AccessObject));

   /*** If an effect has been specified, initialise the timer ***/

   if (Self->Effect) {
      if (Self->Effect IS EFX_TYPIST);
      else if (Self->Effect IS EFX_FLASH) {
         Self->prvFlash.Red   = Self->Font->ColourRGB->Red;
         Self->prvFlash.Green = Self->Font->ColourRGB->Green;
         Self->prvFlash.Blue  = Self->Font->ColourRGB->Blue;
         Self->Font->ColourRGB->Red   = 0;
         Self->Font->ColourRGB->Green = 0;
         Self->Font->ColourRGB->Blue  = 0;
      }

      if ((!Self->EffectDelay) AND (!Self->Frame) AND (!Self->Frames)) {
         /*** Start the effect now if their are no predetermined triggers ***/
         Self->EffectDelay = -1; /* No delay */
         Self->EffectActive = 1; /* Value of 1 means that the effect is active */
      }

      SubscribeTimer(Self->Head.UniqueID, 20);
   }
   else if (Self->Flags & TXF_EDIT) {
      SubscribeTimer(Self->Head.UniqueID, 100); /* Flash the cursor via the timer */
   }

   /* In command line mode, the cursor is placed at the end of any existing
   ** text on initialisation.
   */

   if (Self->LineLimit IS 1) {
      if (Self->Array[0].String) Self->CursorColumn = Self->Array[0].Length;
   }

   CalculateHScroll(Self);
   CalculateVScroll(Self);

   return(ERR_Okay);
}


/*****************************************************************************

###ACTION###
Name:  LostFocus
Short: Calling this action will deactivate keyboard input.
###END###

*****************************************************************************/

static ERROR TEXT_LostFocus(struct Text *Self, APTR Void)
{
   OBJECTPTR input;
   UBYTE redraw;

   DPrintF("~LostFocus()","[Text:%d]", Self->Head.UniqueID);

   if (Self->KeyboardChannel IS TRUE) {
      if (AccessObject(Self->InputID, 5000, &input) IS ERR_Okay) {
         UnsubscribeChannel(input, Self->Head.UniqueID);
         Self->KeyboardChannel = FALSE;
         RemoveCursor(Self);

         ReleaseObject(input);
      }
      else ObjectError(Self, ERH_LostFocus, ERR_AccessObject);
   }

   /*** When a simple input line loses the focus, all selections are deselected ***/

   redraw = FALSE;
   if (Self->LineLimit IS 1) {
      if (Self->Flags & TXF_AREASELECTED) { redraw = TRUE; Self->Flags &= ~TXF_AREASELECTED; }
      if (Self->XPosition) { Self->XPosition = 0; redraw = TRUE; }
      if (redraw) Redraw(Self);
   }

   /* If FOCUSACTIVATE is set, we need to call the Activate action
   ** when the focus is lost.
   */

   if (Self->Flags & TXF_FOCUSACTIVATE) {
      Action(AC_Activate, Self, NULL);
   }

   StepBack();
   return(ERR_Okay);
}

/*****************************************************************************
** Text: NewObject
*/

static ERROR TEXT_NewObject(struct Text *Self, APTR Void)
{
   if (NewObject(ID_FONT, NF_CHILD, &Self->Font, NULL) IS ERR_Okay) {
      SetField(Self->Font, FID_Face, FT_STRING, "sans serif");
      Self->Highlight.Red   = glHighlight.Red;
      Self->Highlight.Green = glHighlight.Green;
      Self->Highlight.Blue  = glHighlight.Blue;
      Self->CursorRGB.Red   = 100;
      Self->CursorRGB.Green = 100;
      Self->CursorRGB.Blue  = 200;
      Self->MaxLines        = 50;
      Self->HistorySize     = 20;
      Self->CursorWidth     = 1;
      Self->CharLimit       = 4096; /* Maximum number of characters per line */
      Self->LineLimit       = 0x7fffffff;
      Self->Duration        = 1000;
      Self->Visible         = TRUE;
      if (AllocMemory(sizeof(struct TextLine) * Self->MaxLines, MEM_DATA, (APTR *)&Self->Array, NULL) IS ERR_Okay) {
         return(ERR_Okay);
      }
      else return(ObjectError(Self, ERH_NewObject, ERR_AllocMemory));
   }
   else return(ObjectError(Self, ERH_NewObject, ERR_NewObject));
}

/*****************************************************************************
** Text: NewOwner
*/

static ERROR TEXT_NewOwner(struct Text *Self, struct acNewOwner *Args)
{
   OBJECTID OwnerID;

   if (!(Self->Head.ObjectFlags & NF_INITIALISED)) {
      OwnerID = Args->NewOwnerID;
      while ((OwnerID) AND (GetClassID(OwnerID) != ID_RENDER)) {
         OwnerID = GetOwnerID(OwnerID);
      }
      if (OwnerID) Self->DrawableID = OwnerID;
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Text: OpeningTag
*/

static ERROR TEXT_OpeningTag(struct Text *Self, APTR Void)
{
   Self->Tag++;
   Self->NoUpdate++;
   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:  ReplaceLine()
Short: Replaces the content of any text line.

###DESCRIPTION###
<p>Any line within a text object can be replaced with new information by
using the ReplaceLine() method.  You need to provide the text string that
you wish to use, the number of the line that will be replaced, and the
length of the text string.</p>

<p>If you set the String argument to NULL, then an empty string will replace
the line number.  If the Length is set to -1, then the length of the new string
will be calculated by counting the amount of characters in the String
argument.</p>

<p>If the new line content is visible within the text object's associated
drawable, that region of the drawable will be redrawn so that the new line
content is displayed.</p>

###INPUT###
STRING/String: The text data that you want to use in replacing the line.
LONG/Line:   The line number that will be replaced.
LONG/Length: The length of the String in bytes.

###RESULT###
ERR_Okay:        Success.
ERR_Args:        Invalid arguments were specified.
ERR_AllocMemory: The memory required to add the text string to the list was
                 unavailable.

###SEE ALSO###
Method: AddLine

###END###

*****************************************************************************/

static ERROR TEXT_ReplaceLine(struct Text *Self, struct mtxReplaceLine *Args)
{
   if (!Args) return(ObjectError(Self, MTX_ReplaceLine, ERR_Args));

   return(ReplaceLine(Self, Args->String, Args->Line, Args->Length));
}

/*****************************************************************************

###METHOD###
Name:  ReplaceText
Short: Searches for and replaces keywords.

###DESCRIPTION###
<p>This method is used to replace matching keywords within a text object
in a single function call.  It is suitable only for mass replacement
of keywords.  If single keyword replacement is required, you should combine
use of the FindText() method and DataChannel action instead of using the
ReplaceText() method.</p>

<p>Optional flags that alter the behaviour of this method are listed in the
following table:</p>

<table border="1" bordercolordark="#000000" cellspacing="0" cellpadding="2" bordercolorlight="#000000" width="80%" align="center">
<tr><td class="tableheader" bgcolor="#eeeeee">Flag</td><td class="tableheader" bgcolor="#eeeeee">Description</td></tr>
<tr><td width="10%">FTF_CASE</td><td>Use case sensitivity when looking for the keyword.</td></tr>
<tr><td width="10%">FTF_FROMCURSOR</td><td>Start the search from the current cursor position and stop at the end of the text data.</td></tr>
<tr><td width="10%">FTF_MOVECURSOR</td><td>Move the cursor to the position of the keyword that was found last.</td></tr>
<tr><td width="10%">FTF_SELECTION</td><td>Search only within the current text selection.  If no text has been selected, do not perform the replace operation.</td></tr>
</table>

###INPUT###
STRING/Find:    The keyword that you want to search for.
STRING/Replace: The text that will replace the keyword.  If NULL, all instances of the keyword will be deleted.
LONG/Flags:     Optional flags.

###RESULT###
ERR_Okay:   The keyword was found and replaced.
ERR_Args:   Invalid arguments were specified.
ERR_Search: The keyword was not found.

###END###

*****************************************************************************/

static ERROR TEXT_ReplaceText(struct Text *Self, struct mtxReplaceText *Args)
{
   STRING str;
   LONG row, column, endrow, endcolumn, pos, i, lastrow, lastcolumn, strcase;
   ERROR error;

   if ((!Args) OR (!Args->Find) OR (!*Args->Find)) {
      return(ObjectError(Self, MTX_ReplaceText, ERR_Args));
   }

   if (Self->AmtLines < 1) return(ERR_Search);

   if (Args->Flags & FTF_SELECTION) {
      if (Self->Flags & TXF_AREASELECTED) {
         GetSelectedArea(Self, &row, &column, &endrow, &endcolumn);
      }
      else return(ERR_Search);
   }
   else if (Args->Flags & FTF_FROMCURSOR) {
      row    = Self->CursorRow;
      column = Self->CursorColumn;
      endrow = Self->AmtLines - 1;
      endcolumn = Self->Array[endrow].Length;
   }
   else {
      row       = 0;
      column    = 0;
      endrow    = Self->AmtLines - 1;
      endcolumn = Self->Array[endrow].Length;
   }

   if ((row >= Self->AmtLines) OR (row < 0)) return(ERR_Search);
   if (column > Self->Array[row].Length) column = Self->Array[row].Length;

   if (Args->Flags & FTF_CASE) strcase = TRUE;
   else strcase = FALSE;

   //DPrintF("@ReplaceText:","[Text] Find: \"%s\" Replace: \"%s\" FROM %dx%d TO %dx%d", Args->Find, Args->Replace, column, row, endcolumn, endrow);

   lastrow = 0;
   lastcolumn = 0;
   error = ERR_Search;
   while (row <= endrow) {
      /*** Search the string ***/

      str = Self->Array[row].String;
      if (Args->Flags & FTF_CASE) {
         for (pos=column; pos < Self->Array[row].Length; pos++) {
            for (i=0; Args->Find[i]; i++) {
               if (str[pos+i] != Args->Find[i]) break;
            }
            if (!Args->Find[i]) break; /* Keyword found */
         }
      }
      else for (pos=column; pos < Self->Array[row].Length; pos++) {
         for (i=0; Args->Find[i]; i++) {
            if (UCase(str[pos+i]) != UCase(Args->Find[i])) break;
         }
         if (!Args->Find[i]) break; /* Keyword found */
      }

      if (pos < Self->Array[row].Length) {
         /*** The keyword was found at the position indicated by pos ***/

         if ((row < endrow) OR ((row IS endrow) AND (pos < endcolumn))) {
            lastrow = row;
            lastcolumn = column;

            /*** Replace all occurances of the keyword on this line ***/

            if (StrReplace(Self->Array[row].String, Args->Find, Args->Replace, &str, strcase) IS ERR_Okay) {
                /*** Remove the existing string and associated tags ***/
               FreeMemory(Self->Array[row].String);
               if (Self->Array[row].Tags) FreeMemory(Self->Array[row].Tags);

               /*** Insert the new string ***/
               for (i=0; str[i]; i++);
               Self->Array[row].String = str;
               Self->Array[row].Tags   = NULL;
               Self->Array[row].Length = i;
               Self->Array[row].PixelLength = CalculateWidth(Self, str, i);
               error = ERR_Okay;
            }
            else DPrintF("@ReplaceText:","StrReplace() failed.");
         }
      }
      column = 0;
      row++;
   }

   if (error != ERR_Search) {
      CalculateHScroll(Self);
      CalculateVScroll(Self);
      if (Args->Flags & FTF_MOVECURSOR) MoveCursor(Self, lastrow, lastcolumn);
      ActionMsg(AC_Draw, Self->DrawableID, NULL);
   }

   return(error);
}

/*****************************************************************************

###ACTION###
Name:  SaveToObject()
Short: Use this action to save edited information as a text file.
###END###

*****************************************************************************/

static ERROR TEXT_SaveToObject(struct Text *Self, struct acSaveToObject *Args)
{
   struct acWrite write;
   OBJECTPTR Object;
   LONG i;

   DPrintF("SaveToObject()","Text: %d, Destination: %d, Lines: %d", Self->Head.UniqueID, Args->DestinationID, Self->AmtLines);

   if (!Args->DestinationID) return(ObjectError(Self, ERH_SaveToObject, ERR_Args));

   if ((Self->Array) AND (Self->AmtLines > 0)) {
      if ((Self->AmtLines IS 1) AND (Self->Array[0].Length < 1)) return(ERR_Okay);

      if (AccessObject(Args->DestinationID, 5000, &Object) IS ERR_Okay) {
         for (i=0; i < Self->AmtLines; i++) {
            /*** Output line ***/
            if (Self->Array[i].Length > 0) {
               write.Buffer = Self->Array[i].String;
               write.Length = Self->Array[i].Length;
               Action(AC_Write, Object, &write);
            }
            /*** Output return code ***/
            write.Length = 1;
            write.Buffer = "\n";
            Action(AC_Write, Object, &write);
         }
         ReleaseObject(Object);
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  ScrollToPoint
Short: Scrolls a text object's graphical content.
###END###

*****************************************************************************/

static ERROR TEXT_ScrollToPoint(struct Text *Self, struct acScrollToPoint *Args)
{
   struct mtMoveContent move;
   OBJECTPTR drawable;
   LONG oldx, oldy, xchange, ychange, x, y;

   if (!Args) return(ObjectError(Self, ERH_ScrollToPoint, ERR_Args));

   if ((Args->XCoord IS Self->XPosition) AND (Args->YCoord IS Self->YPosition)) return(ERR_Okay);

   if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
      if (Args->Flags & STF_X) x = -Args->XCoord;
      else x = Self->XPosition;

      if (Args->Flags & STF_Y) y = -Args->YCoord;
      else y = Self->YPosition;

      oldx = Self->XPosition;
      oldy = Self->YPosition;

      Self->XPosition = x;
      Self->YPosition = y;

      xchange = Self->XPosition - oldx;
      ychange = Self->YPosition - oldy;

      if (Self->Flags & TXF_PRESERVEBKGD) {
         Redraw(Self);
      }
      else if ((xchange >= Self->AreaWidth-10) OR (ychange >= Self->AreaHeight-10) OR
          (xchange <= -Self->AreaWidth+10) OR (ychange <= -Self->AreaHeight+10)) {
         Redraw(Self);
      }
      else {
         /*** Set horizontal movement parameters ***/

         if (x < oldx) {
            move.XCoord = Self->AreaX - xchange;
            move.Width  = Self->AreaWidth;
            move.XDest  = Self->AreaX;
         }
         else {
            move.XCoord = Self->AreaX - xchange;
            move.Width  = Self->AreaWidth;
            move.XDest  = Self->AreaX;
         }

         /*** Set vertical movement parameters ***/

         if (y < oldy) {
            move.YCoord = Self->AreaY - ychange;
            move.Height = Self->AreaHeight;
            move.YDest  = Self->AreaY;
         }
         else {
            move.YCoord = Self->AreaY - ychange;
            move.Height = Self->AreaHeight;
            move.YDest  = Self->AreaY;
         }

         move.ClipLeft   = Self->AreaX;
         move.ClipTop    = Self->AreaY;
         move.ClipRight  = Self->AreaX + Self->AreaWidth;
         move.ClipBottom = Self->AreaY + Self->AreaHeight;

         move.Flags  = MCF_EXPOSE;
         Action(MT_MoveContent, drawable, &move);
      }

      ReleaseObject(drawable);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:    SelectTextArea
Short:   Forces a user selection over a specific text area.

###DESCRIPTION###
<p>If you would like to force a user selection over a specific text area,
use this method.  Normally, user selections occur when the user moves a
pointing device over a Text object to highlight an area of text.  By
using this method, you will bypass that procedure by highlighting an area
manually.</p>

<p>The area that you specify will be highlighted as if the user had selected
that area him or herself.  The selection can be cancelled if the user performs
an activity that causes the selection to be removed.</p>

###INPUT###
LONG/Row: The row from which the selection will start.
LONG/Column: The column from which the selection will start.
LONG/EndRow: The number of the row that will terminate the selection.
LONG/EndColumn: The number of the column that will terminate the selection.

###RESULT###
ERR_Okay: The area has been selected successfully.
ERR_Args: Invalid arguments were specified.

###END###

*****************************************************************************/

static ERROR TEXT_SelectTextArea(struct Text *Self, struct mtSelectTextArea *Args)
{
   if ((!Args) OR (Args->Row < 0) OR (Args->Column < 0) OR
       (Args->EndRow < 0) OR (Args->EndColumn < 0)) {
      return(ObjectError(Self, MT_SelectTextArea, ERR_Args));
   }

   DPrintF("SelectArea()","[Text] %dx%d TO %dx%d", Args->Column, Args->Row, Args->EndColumn, Args->EndRow);

   if (Self->AmtLines < 1) {
      DPrintF("SelectArea:","[Text] There is no selectable data present.");
      return(ERR_Okay);
   }

   if (Args->Row < Self->AmtLines) Self->SelectRow = Args->Row;
   else Self->SelectRow = Self->AmtLines - 1;

   if (Args->Column < Self->Array[Self->SelectRow].Length) Self->SelectColumn = Args->Column;
   else Self->SelectColumn = Self->Array[Self->SelectRow].Length;

   if (Args->EndRow < Self->AmtLines) Self->CursorRow = Args->EndRow;
   else Self->CursorRow = Self->AmtLines - 1;

   if (Args->EndColumn < Self->Array[Self->CursorRow].Length) Self->CursorColumn = Args->EndColumn;
   else Self->CursorColumn = Self->Array[Self->CursorRow].Length;

   if ((Self->SelectRow != Self->CursorRow) OR (Self->SelectColumn != Self->CursorColumn)) {
      Self->Flags |= TXF_AREASELECTED;
   }
   else {
      DPrintF("SelectArea:","No text was selected.");
      Self->Flags &= ~TXF_AREASELECTED;
   }

   DrawLines(Self, Self->SelectRow, Self->CursorRow - Self->SelectRow + 1);
   ViewSelection(Self);

   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:  SetFont
Short: Allows you to change the font face, style and size after initialisation.
###END###

*****************************************************************************/

static ERROR TEXT_SetFont(struct Text *Self, struct mtxSetFont *Args)
{
   struct Font *font;
   LONG flags, i;

   if ((!Args) OR (!Args->Face)) return(ObjectError(Self, MTX_SetFont, ERR_Args));

   flags = FTF_DIRECTXY;
   if (Self->Flags & TXF_WORDWRAP) flags |= FTF_WORDWRAP;

   if (CreateObject(ID_FONT, NF_CHILD, &font, NULL,
      FID_Face|TSTRING, Args->Face,
      FID_Flags|TLONG,  flags,
      TAGEND) IS ERR_Okay) {

      if (Self->Font)       Action(AC_Free, Self->Font, NULL);
      if (Self->BoldFont)   Action(AC_Free, Self->BoldFont, NULL);
      if (Self->ItalicFont) Action(AC_Free, Self->ItalicFont, NULL);
      if (Self->BIFont)     Action(AC_Free, Self->BIFont, NULL);

      Self->Font       = font;
      Self->BoldFont   = NULL;
      Self->ItalicFont = NULL;
      Self->BIFont     = NULL;

      /*** Recalculate the pixel width of each line ***/

      for (i=0; i < Self->AmtLines; i++) {
         Self->Array[i].PixelLength = CalculateWidth(Self, Self->Array[i].String, Self->Array[i].Length);
      }

      Self->CursorRow    = 0;
      Self->CursorColumn = 0;
      Self->SelectRow    = 0;
      Self->SelectColumn = 0;
      Self->XPosition    = 0;
      Self->YPosition    = 0;
      Self->Flags &= ~TXF_AREASELECTED;

      Redraw(Self);
      CalculateHScroll(Self);
      CalculateVScroll(Self);

      return(ERR_Okay);
   }
   else return(ERR_CreateObject);
}

/*****************************************************************************
** Text: Show
*/

static ERROR TEXT_Show(struct Text *Self, APTR Void)
{
   if (Self->Visible IS FALSE) {
      Self->Visible = TRUE;
      if (Self->Head.ObjectFlags & NF_INITIALISED) Redraw(Self);
      return(ERR_Okay);
   }
   else return(ERR_Okay|ERF_Notified);
}

/*****************************************************************************
** Text: Timer
*/

static ERROR TEXT_Timer(struct Text *Self, struct acTimer *Args)
{
   OBJECTPTR drawable;
   LONG len, frame, i;

   if (Self->Flags & TXF_EDIT) {
      UBYTE one, two;
      one = (Self->CursorFlash % CURSOR_RATE) < (CURSOR_RATE>>1);
      Self->CursorFlash += Args->TimeElapsed;
      two = (Self->CursorFlash % CURSOR_RATE) < (CURSOR_RATE>>1);

      if (Self->KeyboardChannel) {
         if (Self->LineLimit IS 1) ViewCursor(Self);

         if (one != two) RedrawCursor(Self, TRUE);
      }
      return(ERR_Okay);
   }

   if (Self->Effect) {
      if (Self->AmtLines < 1) {
         Self->Effect = NULL;
         UnsubscribeTimer(Self->Head.UniqueID);
         return(ERR_Okay);
      }

      if (Self->Duration < 2) Self->Duration = 2; /* Duration must be greater than zero */

      if (Self->EffectDelay > 0) {
         if (Args) Self->EffectDelay -= Args->TimeElapsed;
         if (Self->EffectDelay > 0) return(ERR_Okay);

         Self->EffectActive = 1; /* Value of 1 means that the effect is active */
         Self->EffectDelay = -1; /* No delay */
         SubscribeTimer(Self->Head.UniqueID, 20); /* Resubscribe to reset the total elapsed time */
         return(ERR_Okay);
      }
      else if ((Self->Frame) OR (Self->Frames)) {
         if ((!Self->EffectActive) OR (Self->EffectActive IS 2)) {
            if (AccessObject(Self->DrawableID, 0, &drawable) IS ERR_Okay) {
               GetField(drawable, FID_Frame, FT_LONG, &frame);
               ReleaseObject(drawable);

               /*** Reset the effect if the animation has looped back to frame zero ***/

               if (Self->EffectActive IS 2) {
                  if (frame < Self->CurrentFrame) {
                     Self->EffectActive = 0; /* Value of 0 means that the effect is inactive and waiting to be restarted */
                     goto reseteffect;
                  }
               }

               if (!Self->EffectActive) {
                  if (Self->Frame) {
                     if (frame IS Self->Frame) {
                        Self->EffectActive = 1; /* Value of 1 means that the effect is active */
                        goto reseteffect;
                     }
                  }
                  else {
                     if ((frame <= Self->TotalFrames) AND (Self->Frames[frame] >= 1)) {
                        Self->EffectActive = 1;
                        goto reseteffect;
                     }
                  }
               }

               Self->CurrentFrame = frame;
            }
            else return(ObjectError(Self, ERH_Timer, ERR_AccessObject));
         }

         /*** Return if the activity variable does not have a value of 1 ***/

         if ((!Self->EffectActive) OR (Self->EffectActive IS 2)) return(ERR_Okay);
      }

      if (Self->Effect IS EFX_TYPIST) {
         /*** Typist effect type ***/

         len = 0;
         for (i=0; i < Self->AmtLines; i++) {
            len += Self->Array[i].Length;
         }

         Self->CharCount = (Args->TotalElapsed * len) / Self->Duration;

         if (Self->CharCount >= len) {
            if (!(Self->Flags & TXF_EFFECTLOOP)) UnsubscribeTimer(Self->Head.UniqueID);
            Self->CharCount = 0;
            Self->EffectActive = 2;
         }

         if (!(Self->Flags & TXF_NOFXDRAW)) {
            Redraw(Self);
         }
      }
      else if ((Self->Effect IS EFX_FLASH) OR (Self->Effect IS EFX_FLASHDOWN)) {
         if (Self->Effect IS EFX_FLASH) {
            /*** Flash text to white ***/

            Self->FlashCount = (Args->TotalElapsed * 255) / (Self->Duration>>1);
            if (Self->FlashCount > 255) Self->FlashCount = 255;

            Self->Font->ColourRGB->Red   = Self->FlashCount;
            Self->Font->ColourRGB->Green = Self->FlashCount;
            Self->Font->ColourRGB->Blue  = Self->FlashCount;

            if (Self->FlashCount >= 255) Self->Effect = EFX_FLASHDOWN;
         }
         else {
            /*** Flash text from white to its natural colour ***/

            Self->FlashCount = ((Self->Duration - Args->TotalElapsed)<<8) / (Self->Duration>>1);
            if (Self->FlashCount < 0) Self->FlashCount = 0;

            Self->Font->ColourRGB->Red   = Self->prvFlash.Red   + (((255 - Self->prvFlash.Red) * Self->FlashCount)>>8);
            Self->Font->ColourRGB->Green = Self->prvFlash.Green + (((255 - Self->prvFlash.Green) * Self->FlashCount)>>8);
            Self->Font->ColourRGB->Blue  = Self->prvFlash.Blue  + (((255 - Self->prvFlash.Blue) * Self->FlashCount)>>8);

            if (!Self->FlashCount) {
               if (!(Self->Flags & TXF_EFFECTLOOP)) UnsubscribeTimer(Self->Head.UniqueID);
               Self->Effect = EFX_FLASH;
               Self->EffectActive = 2;
            }
         }

         if (!(Self->Flags & TXF_NOFXDRAW)) {
            Redraw(Self);
         }
      }
      else {
         /*** Unknown effect type ***/
         Self->Effect = NULL;
         UnsubscribeTimer(Self->Head.UniqueID);
      }
   }

   return(ERR_Okay);

reseteffect:
   if ((Self->Effect IS EFX_FLASH) OR (Self->Effect IS EFX_FLASHDOWN)) {
      Self->Font->ColourRGB->Red   = 0;
      Self->Font->ColourRGB->Green = 0;
      Self->Font->ColourRGB->Blue  = 0;
   }
   else if (Self->Effect IS EFX_TYPIST) {
      Self->CharCount = 0;
   }

   SubscribeTimer(Self->Head.UniqueID, 20); /* Resubscribe to reset the total elapsed time */
   ActionMsg(AC_Draw, Self->DrawableID, NULL);
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  UserClick
Short: Supported for responding to user clicks (authentic or faked).
###END###

*****************************************************************************/

static ERROR TEXT_UserClick(struct Text *Self, struct acUserClick *Args)
{
   struct acDraw draw;
   struct mtConvertFontCoords convert;
   STRING str;
   LONG rowcount, i, clickrow, clickcol;
   UBYTE outofbounds;

   if (!Args) return(ObjectError(Self, ERH_UserClick, ERR_Args));

   if (!(Args->ButtonFlags & JD_LMB)) return(ERR_Okay);

   if (!(Self->Flags & (TXF_EDIT|TXF_SINGLESELECT|TXF_MULTISELECT))) return(ERR_Okay);

   DPrintF("~UserClick()","[Text:%d]", Self->Head.UniqueID);

   Self->CursorFlash = 0;
   outofbounds = FALSE;
   clickrow = 0;
   clickcol = 0;

   /*** Determine the row that was clicked ***/

   if (Self->AmtLines > 0) {
      clickrow = (Args->YCoord - Self->TopMargin - Self->YPosition) / Self->Font->LineSpacing;
      if (clickrow >= Self->AmtLines) {
         clickrow = Self->AmtLines - 1;
         outofbounds = TRUE;
      }
   }

   /*** Determine the column that was clicked ***/

   if (clickrow < Self->AmtLines) {
      if (Self->Array[clickrow].String) {
         if (Self->Flags & TXF_SECRET) {
            BYTE buffer[Self->Array[clickrow].Length+1];
            for (i=0; i < Self->Array[clickrow].Length; i++) buffer[i] = '*';
            buffer[i+1] = 0;
            convert.String = buffer;
            convert.XCoord = Args->XCoord - Self->AreaX - Self->LeftMargin - Self->XPosition;
            convert.YCoord = 0;
            if (Action(MT_ConvertFontCoords, Self->Font, &convert) IS ERR_Okay) {
               clickcol = convert.ByteColumn;
            }
         }
         else {
            convert.String = Self->Array[clickrow].String;
            convert.XCoord = Args->XCoord - Self->AreaX - Self->LeftMargin - Self->XPosition;
            convert.YCoord = 0;
            if (Action(MT_ConvertFontCoords, Self->Font, &convert) IS ERR_Okay) {
               clickcol = convert.ByteColumn;
            }
         }
      }
   }

   /*** If there is an old area selection, clear it ***/

   if (Self->Flags & TXF_AREASELECTED) {
      Self->Flags &= ~TXF_AREASELECTED;

      draw.XCoord = Self->AreaX;
      draw.Width  = Self->AreaWidth;
      if (Self->CursorRow < Self->SelectRow) {
         draw.YCoord = RowCoordinate(Self, Self->CursorRow);
         rowcount = Self->SelectRow - Self->CursorRow + 1;
      }
      else {
         draw.YCoord = RowCoordinate(Self, Self->SelectRow);
         rowcount = Self->CursorRow - Self->SelectRow + 1;
      }
      draw.Height = rowcount * Self->Font->LineSpacing;

      Self->NoCursor++;
      ActionMsg(AC_Draw, Self->DrawableID, &draw);
      Self->NoCursor--;
   }
   else if (Self->Flags & TXF_EDIT) {
      RemoveCursor(Self);
   }

   Self->SelectRow    = clickrow;
   Self->SelectColumn = clickcol;
   Self->ClickHeld    = TRUE;

   /*** Return if we are NOT in edit mode and the position of the click was out of bounds ***/

   if ((outofbounds) AND (!(Self->Flags & TXF_EDIT))) {
      StepBack();
      return(ERR_Okay);
   }

   /*** Return if the row and column values will remain unchanged ***/

   if (((clickcol IS Self->CursorColumn) AND (clickrow IS Self->CursorRow)) OR
       (Self->AmtLines < 1)) {
      if (!(Args->ButtonFlags & JD_DOUBLECLICK)) {
         RedrawCursor(Self, TRUE);
         StepBack();
         return(ERR_Okay);
      }
   }

   Self->CursorRow    = clickrow;
   Self->CursorColumn = clickcol;

   /*** For double-clicks, highlight the word next to the cursor ***/

   if ((Args->ButtonFlags & JD_DOUBLECLICK) AND (!(Self->Flags & TXF_SECRET))) {
      /*** Scan back to find the start of the word ***/

      str = Self->Array[Self->CursorRow].String;
      for (i=Self->CursorColumn; i > 0; i--) {
         if (str[i-1] <= 47) break;
         if ((str[i-1] >= 58) AND (str[i-1] <= 64)) break;
         if ((str[i-1] >= 91) AND (str[i-1] <= 96)) break;
         if ((str[i-1] >= 123) AND (str[i-1] <= 127)) break;
      }

      Self->SelectColumn = i;

      /*** Scan forward to find the end of the word ***/

      for (i=Self->CursorColumn; i < Self->Array[Self->CursorRow].Length; i++) {
         if (str[i] <= 47) break;
         if ((str[i] >= 58) AND (str[i] <= 64)) break;
         if ((str[i] >= 91) AND (str[i] <= 96)) break;
         if ((str[i] >= 123) AND (str[i] <= 127)) break;
      }

      Self->CursorColumn = i;

      Self->SelectRow = Self->CursorRow;

      if (Self->SelectColumn != Self->CursorColumn) {
         Self->Flags |= TXF_AREASELECTED;
         RedrawLine(Self, Self->CursorRow);
      }
      else RedrawCursor(Self, TRUE);
   }
   else RedrawCursor(Self, TRUE);

   ViewCursor(Self);

   StepBack();
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  UserClickRelease
Short: Supported for the purpose of responding to user click releases.
###END###

*****************************************************************************/

static ERROR TEXT_UserClickRelease(struct Text *Self, struct acUserClickRelease *Args)
{
   if (!(Self->Flags & (TXF_EDIT|TXF_SINGLESELECT|TXF_MULTISELECT))) return(ERR_Okay);

   DPrintF("UserClickRelease()","[Text:%d]", Self->Head.UniqueID);

   if (Args->ButtonFlags & JD_LMB) {
      Self->ClickHeld = FALSE;
      if ((Self->SelectRow != Self->CursorRow) OR (Self->SelectColumn != Self->CursorColumn)) {
         Self->Flags |= TXF_AREASELECTED;
      }
   }

   StepBack();
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  UserMovement
Short: Supported for the purpose of responding to user movement.
###END###

*****************************************************************************/

static ERROR TEXT_UserMovement(struct Text *Self, struct acUserMovement *Args)
{
   struct mtConvertFontCoords convert;
   LONG oldrow, oldcolumn;

   if ((Self->ClickHeld IS FALSE) OR (!Args)) return(ERR_Okay);

   if (Self->AmtLines < 1) return(ERR_Okay);

   if (!(Self->Flags & (TXF_EDIT|TXF_SINGLESELECT|TXF_MULTISELECT))) return(ERR_Okay);

   if (Self->Flags & TXF_SECRET) return(ERR_Okay);

   oldrow = Self->CursorRow;
   oldcolumn = Self->CursorColumn;

   /*** Calculate the cursor row ***/

   Self->CursorRow = (Args->YCoord - Self->TopMargin - Self->YPosition) / Self->Font->LineSpacing;

   if (Self->CursorRow < 0) Self->CursorRow = 0;
   if (Self->CursorRow >= Self->AmtLines) Self->CursorRow = Self->AmtLines - 1;

   /*** Calculate the cursor column ***/

   Self->CursorColumn = 0;
   if (Self->Array[Self->CursorRow].String) {
      convert.String = Self->Array[Self->CursorRow].String;
      if ((convert.XCoord = Args->XCoord - Self->AreaX - Self->LeftMargin - Self->XPosition) < 0) convert.XCoord = 0;
      convert.YCoord = 0;
      if (Action(MT_ConvertFontCoords, Self->Font, &convert) IS ERR_Okay) {
         Self->CursorColumn = convert.ByteColumn;
      }
   }

   if ((Self->CursorRow != oldrow) OR (Self->CursorColumn != oldcolumn)) {

      /*** Set the AREASELECTED flag if an area has been highlighted by the user ***/

      if ((Self->SelectRow != Self->CursorRow) OR (Self->SelectColumn != Self->CursorColumn)) {
         Self->Flags |= TXF_AREASELECTED;
      }

      if (Self->CursorRow < oldrow) DrawLines(Self, Self->CursorRow, oldrow - Self->CursorRow + 1);
      else DrawLines(Self, oldrow, Self->CursorRow - oldrow + 1);
   }

   ViewCursor(Self);

   return(ERR_Okay);
}

/****************************************************************************/

#include "fields.c"
#include "functions.c"

